"""Extra widgets for Qt."""

from loguru import logger

__version__ = "0.1.4"
__author__ = "Lukasz G. Migas"
__email__ = "lukas.migas@yahoo.com"
__issue_url__ = "https://github.com/lukasz-migas/qtextra-issues/issues"
__project_url__ = "https://lukasz-migas.com/qtextra"
logger.disable("qtextra")



"""Base dialog."""

from __future__ import annotations

import typing as ty
from contextlib import contextmanager, suppress

import numpy as np
from loguru import logger
from qtpy.QtCore import (  # type: ignore[attr-defined]
    QEasingCurve,
    QObject,
    QPoint,
    QPropertyAnimation,
    QRect,
    QSize,
    Qt,
    QTimer,
    Signal,
)
from qtpy.QtGui import QCloseEvent, QCursor, QGuiApplication, QKeyEvent, QMouseEvent, QResizeEvent
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QFrame,
    QGraphicsOpacityEffect,
    QHBoxLayout,
    QLabel,
    QLayout,
    QVBoxLayout,
    QWidget,
)

import qtextra.helpers as hp
from qtextra.config import EVENTS
from qtextra.mixins import CloseMixin, ConfigMixin, DocumentationMixin, IndicatorMixin, TimerMixin

if ty.TYPE_CHECKING:
    from qtextra.widgets.qt_button_icon import QtExpandButton


class ScreenManager:
    """Simple class that handles multi-screen logic."""

    def __init__(self) -> None:
        from qtpy.QtWidgets import QApplication

        self.screens = QApplication.screens()
        self.widths = [screen.geometry().width() for screen in self.screens]
        self.width = sum(self.widths)
        self.heights = [screen.geometry().height() for screen in self.screens]
        self.height = sum(self.heights)

    def get_minimum_size(self, width: int, height: int) -> tuple[int, int]:
        """Get size that is suggested for current screen sizes."""
        self.widths.append(width)
        self.heights.append(height)
        return np.min(self.widths), np.min(self.heights)

    def verify_position(self, point: QPoint, width: int, height: int) -> QPoint:
        """Verify widget position is within the available geometry."""
        x_left, y_top = point.x(), point.y()
        # verify position horizontally
        if x_left < 0:
            x_left = 0
        x_right = x_left + width
        if x_right > self.width:
            x_right = self.width
            x_left = x_right - width
        # verify position vertically
        if y_top < 0:
            y_top = 0
        y_bottom = y_top - height
        if y_bottom > self.height:
            y_bottom = self.height
            y_top = y_bottom - height
        return QPoint(x_left, y_top)


class DialogMixin:
    """Mixin class for dialogs."""

    def show_on_mouse(self, show: bool = True) -> None:
        """Show popup dialog in the center of mouse cursor position."""
        hp.show_on_mouse(self, show)

    def show_right_of_mouse(self, show: bool = True) -> None:
        """Show popup dialog on the right hand side of the mouse cursor position."""
        hp.show_right_of_mouse(self, show)

    def show_left_of_mouse(self, show: bool = True) -> None:
        """Show popup dialog on the left hand side of the mouse cursor position."""
        hp.show_left_of_mouse(self, show)

    def show_above_mouse(self, show: bool = True, x_offset: int = 0, y_offset: int = 0) -> None:
        """Show popup dialog above the mouse cursor position."""
        hp.show_above_mouse(self, show, x_offset=x_offset, y_offset=y_offset)

    def show_below_mouse(self, show: bool = True, x_offset: int = 0, y_offset: int = -0) -> None:
        """Show popup dialog above the mouse cursor position."""
        hp.show_below_mouse(self, show, x_offset=x_offset, y_offset=y_offset)

    def show_in_center_of_screen(self, show: bool = True) -> None:
        """Show dialog in the center of the widget."""
        hp.show_in_center_of_screen(self, show)

    def show_in_center_of_widget(self, widget: QWidget, show: bool = True) -> None:
        """Show dialog in the center of the widget."""
        hp.show_in_center_of_widget(self, widget, show)

    def show_above_widget(self, parent: QWidget, show: bool = True, x_offset: int = 0, y_offset: int = 0) -> None:
        """Show popup dialog above the widget."""
        hp.show_above_widget(self, parent, show, x_offset=x_offset, y_offset=y_offset)

    def show_below_widget(self, parent: QWidget, show: bool = True, x_offset: int = 0, y_offset: int = 0) -> None:
        """Show popup dialog above the widget."""
        hp.show_below_widget(self, parent, show, x_offset=x_offset, y_offset=y_offset)

    def show_right_of_widget(self, parent: QWidget, show: bool = True, x_offset: int = 0, y_offset: int = 0) -> None:
        """Show popup dialog above the widget."""
        hp.show_right_of_widget(self, parent, show, x_offset=x_offset, y_offset=y_offset)

    def show_left_of_widget(
        self, parent: QObject | None, show: bool = True, x_offset: int = 0, y_offset: int = 0
    ) -> None:
        """Show popup dialog above the widget."""
        hp.show_left_of_widget(self, parent, show, x_offset=x_offset, y_offset=y_offset)

    def move_to(self, position="top", *, win_ratio=0.9, min_length=0) -> None:
        """Move popup to a position relative to the QMainWindow.

        Parameters
        ----------
        position : {str, tuple}, optional
            position in the QMainWindow to show the pop, by default 'top'
            if str: must be one of {'top', 'bottom', 'left', 'right' }
            if tuple: must be length 4 with (left, top, width, height)
        win_ratio : float, optional
            Fraction of the width (for position = top/bottom) or height (for
            position = left/right) of the QMainWindow that the popup will
            occupy.  Only valid when isinstance(position, str).
            by default 0.9
        min_length : int, optional
            Minimum size of the long dimension (width for top/bottom or
            height fort left/right).

        Raises
        ------
        ValueError
            if position is a string and not one of
            {'top', 'bottom', 'left', 'right' }
        """
        if isinstance(position, str):
            window = self.parent().window() if self.parent() else None
            if not window:
                raise ValueError(
                    "Specifying position as a string is only possible if the popup has a parent",
                )
            left = window.pos().x()
            top = window.pos().y()
            if position in ("top", "bottom"):
                width = window.width() * win_ratio
                width = max(width, min_length)
                left += (window.width() - width) / 2
                height = self.sizeHint().height()
                top += 24 if position == "top" else (window.height() - height - 12)
            elif position in ("left", "right"):
                height = window.height() * win_ratio
                height = max(height, min_length)
                # 22 is for the title bar
                top += 22 + (window.height() - height) / 2
                width = self.sizeHint().width()
                left += 12 if position == "left" else (window.width() - width - 12)
            else:
                raise ValueError(
                    'position must be one of ["top", "left", "bottom", "right"]',
                )
        elif isinstance(position, (tuple, list)):
            assert len(position) == 4, "`position` argument must have length 4"
            left, top, width, height = position
        else:
            raise ValueError(
                f"Wrong type of position {position}",
            )

        # necessary for transparent round corners
        self.resize(self.sizeHint())
        # make sure the popup is completely on the screen
        # In Qt ≥5.10 we can use screenAt to know which monitor the mouse is on

        if hasattr(QGuiApplication, "screenAt"):
            screen_geometry: QRect = QGuiApplication.screenAt(QCursor.pos()).geometry()
        else:
            # This widget is deprecated since Qt 5.11
            from qtpy.QtWidgets import QDesktopWidget

            screen_num = QDesktopWidget().screenNumber(QCursor.pos())
            screen_geometry = QGuiApplication.screens()[screen_num].geometry()

        left = max(min(screen_geometry.right() - width, left), screen_geometry.left())
        top = max(min(screen_geometry.bottom() - height, top), screen_geometry.top())
        self.setGeometry(left, top, width, height)

    def move_to_widget(self, widget: QWidget, position: str = "right") -> None:
        """Move tutorial to specified widget."""
        x_pad, y_pad = 5, 5
        size = self.size()
        rect = widget.rect()
        if position == "left":
            x = rect.left() - size.width() - x_pad
            y = rect.center().y() - (size.height() * 0.5)
        elif position == "right":
            x = rect.right() + x_pad
            y = rect.center().y() - (size.height() * 0.5)
        elif position == "top":
            x = rect.center().x() - (size.width() * 0.5)
            y = rect.top() - size.height() - y_pad
        elif position == "bottom":
            x = rect.center().x() - (size.width() * 0.5)
            y = rect.bottom() + y_pad
        pos = widget.mapToGlobal(QPoint(x, y))
        self.move(pos)


class ScreenshotMixin:
    """Mixin class for taking screenshots."""

    @contextmanager
    def run_with_screenshot(self):
        """Must implement."""
        yield

    def _screenshot(self):
        return self.grab().toImage()

    def to_screenshot(self):
        """Get screenshot."""
        from napari._qt.dialogs.screenshot_dialog import ScreenshotDialog

        dialog = ScreenshotDialog(self.screenshot, self, history=[])
        if dialog.exec_():
            pass

    def screenshot(self, path: ty.Optional[str] = None):
        """Take screenshot of the viewer."""
        from napari._qt.utils import QImg2array

        with self.run_with_screenshot():
            img = self._screenshot()
        if path is not None:
            from skimage.io import imsave

            imsave(path, QImg2array(img))
        return QImg2array(img)

    def clipboard(self):
        """Take screenshot af the viewer and put it in the clipboard."""
        from qtextra.widgets.qt_button_clipboard import copy_image_to_clipboard

        with self.run_with_screenshot():
            img = self._screenshot()
        copy_image_to_clipboard(img)
        hp.add_flash_animation(self)

    def _get_save_screenshot_menu(self):
        """Get normalization menu."""
        menu = hp.make_menu(self)
        menu_save = hp.make_menu_item(self, "Save screenshot to file...", menu=menu)
        menu_save.triggered.connect(self.to_screenshot)
        menu_clip = hp.make_menu_item(self, "Copy screenshot to clipboard", menu=menu)
        menu_clip.triggered.connect(self.clipboard)
        return menu


class QtBase(ConfigMixin, DocumentationMixin, IndicatorMixin, TimerMixin, ScreenshotMixin):
    """Mixin class with common functionality for Dialogs and Tabs."""

    _is_init = False
    _main_layout = None
    _title = ""

    DELAY_CONNECTION: ty.ClassVar[bool] = False
    setLayout: ty.Callable[[QLayout], None]

    def __init__(self, parent: QWidget | None = None, title: str = "", delay: bool = False):
        if self._is_init:
            return

        self.logger = logger.bind(src=self.__class__.__name__)
        # Qt stuff
        if hasattr(self, "setWindowTitle"):
            self.setWindowTitle(QApplication.translate(str(self), self._title or title, None, -1))
        if hasattr(self, "setAttribute"):
            self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        # Own attributes
        self._parent = parent
        # Make interface
        self.make_gui()
        # Update values
        self.on_set_from_config()
        # Connect signals
        if not delay and not self.DELAY_CONNECTION:
            self.connect_events()
        self._is_init = True

    def make_panel(self) -> QLayout:
        """Make panel."""
        raise NotImplementedError("Must implement method")

    def make_gui(self) -> None:
        """Make and arrange main panel."""
        layout = self.make_panel()
        if layout is None:
            raise ValueError("Expected layout")
        if not layout.parent():
            self.setLayout(layout)
        self._main_layout = layout

    def on_apply(self, *args: ty.Any) -> None:
        """Update config."""

    def _on_teardown(self) -> None:
        """Teardown."""

    def connect_events(self, state: bool = True) -> None:
        """Connect events."""

    def closeEvent(self, event: QCloseEvent | None) -> None:
        """Hide rather than close."""
        self._on_teardown()
        self.connect_events(False)
        if hasattr(self, "evt_close"):
            self.evt_close.emit()
        self.close()


class QtTab(QWidget, QtBase):  # type: ignore[misc]
    """Dialog base class."""

    _description: dict | None = None
    _tab_index: dict | None = None

    def __init__(self, parent: QWidget | None, title: str = "Panel"):
        QWidget.__init__(self, parent)
        QtBase.__init__(self, parent, title)

    def _make_html_description(self) -> str:
        """Make nicely formatted description that can be used in tooltip information."""
        if not self._description:
            return ""
        return f"<p style='white-space:pre'><h2>{self._description.get('title', 'Panel')}</h2></p>"

    def _make_html_metadata(self) -> ty.Tuple[str, str, str]:
        """Make nicely formatted description that can be used to provide help information about widget."""
        if not self._description:
            return "", "", ""
        return (
            self._description.get("title", "Panel"),
            self._description.get("description", ""),
            self._description.get("docs", ""),
        )


class QtDialog(QDialog, DialogMixin, QtBase, CloseMixin):  # type: ignore[misc]
    """Dialog base class."""

    _main_layout: QLayout | None = None

    # events
    evt_resized = Signal()
    evt_hide = Signal()
    evt_close = Signal()

    def __init__(self, parent: QWidget | None = None, title: str = "Dialog", delay: bool = False):
        QDialog.__init__(self, parent)
        self.setWindowFlag(Qt.WindowType.WindowContextHelpButtonHint, False)
        QtBase.__init__(self, parent, title, delay=delay)

        EVENTS.evt_force_exit.connect(self.close)

    def is_valid_(self) -> bool:
        """Check whether object is valid."""
        try:
            self.isVisible()
        except RuntimeError:
            return False
        return True

    def resizeEvent(self, event: QResizeEvent) -> None:  # type: ignore[override]
        """Resize event."""
        self.evt_resized.emit()
        return super().resizeEvent(event)

    def keyPressEvent(self, event: QKeyEvent) -> None:  # type: ignore[override]
        """Close window on return, else pass event through to super class.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            # self.closeEvent(event)
            event.ignore()
        else:
            super().keyPressEvent(event)

    def closeEvent(self, event: QCloseEvent | None = None) -> None:
        """Hide rather than close."""
        if self.HIDE_WHEN_CLOSE:
            self.hide()
            self.clearFocus()
            if hasattr(self, "evt_hide"):
                self.evt_hide.emit()
            event.ignore()
        else:
            self._on_teardown()
            self.connect_events(False)
            if hasattr(self, "evt_close"):
                self.evt_close.emit()
            super().closeEvent(event)

    def close(self) -> bool:
        """Hide dialog rather than delete it."""
        if self.HIDE_WHEN_CLOSE:
            self.hide()
            if hasattr(self, "evt_hide"):
                self.evt_hide.emit()
            self.clearFocus()
            return False
        else:
            return super().close()


class QtFramelessPopup(QtDialog, CloseMixin):  # type: ignore[misc]
    """Frameless dialog."""

    # attributes used to move windows around
    _title_label: QLabel
    _old_window_pos, _title_layout, _move_handle = None, None, None

    def __init__(
        self,
        parent: ty.Optional[QWidget],
        title: str = "",
        position: QPoint | None = None,
        flags: ty.Any = Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.Popup,
        delay: bool = False,
    ):
        super().__init__(parent, title, delay=delay)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)
        self.setWindowFlags(flags)
        if position is not None:
            self.move(position)

    def toggle_persist(self) -> None:
        """Enable persist."""
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.Tool)

    def toggle_popup(self) -> None:
        """Toggle popup to be temporary."""
        self.setWindowFlags(
            Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.Popup
        )

    def _make_title_handle(self, title: str = "") -> QHBoxLayout:
        """Make handle button that helps move the window around."""
        self._title_label = hp.make_eliding_label2(
            self,
            title,
            bold=True,
            alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
            elide=Qt.TextElideMode.ElideRight,
            object_name="window_title",
        )

        layout = hp.make_hbox_layout(spacing=0)
        layout.addWidget(self._title_label, stretch=True)
        layout.addStretch(1)
        self._title_layout = layout
        return layout

    def _make_move_handle(self, title: str = "") -> QHBoxLayout:
        """Make handle button that helps move the window around."""
        self._title_label = hp.make_eliding_label2(
            self,
            title,
            bold=True,
            alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
            elide=Qt.TextElideMode.ElideRight,
            object_name="window_title",
        )
        self._move_handle = hp.make_qta_label(
            self, "move_handle", tooltip="Click here and drag the mouse around to move the window.", normal=True
        )
        self._move_handle.setCursor(Qt.CursorShape.PointingHandCursor)

        layout = hp.make_hbox_layout(spacing=0)
        layout.addWidget(self._title_label, stretch=True)
        layout.addWidget(self._move_handle)
        self._title_layout = layout
        return layout

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Mouse press event."""
        super().mousePressEvent(event)
        # allow movement of the window when user uses right-click and the move handle button does not exist
        if event.button() == Qt.MouseButton.RightButton:  # and self._move_handle is None:
            self._old_window_pos = event.x(), event.y()
        elif self._move_handle is None:
            self._old_window_pos = None
        elif self.childAt(event.pos()) == self._move_handle:
            self._old_window_pos = event.x(), event.y()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """Mouse move event - ensures its possible to move the window to new location."""
        super().mouseMoveEvent(event)
        if self._old_window_pos is not None:
            self.move(event.globalX() - self._old_window_pos[0], event.globalY() - self._old_window_pos[1])

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """Mouse release event."""
        super().mouseReleaseEvent(event)
        self._old_window_pos = None

    def disable_while_open(self, *widgets: QWidget) -> None:
        """Disable widgets while the window is open."""
        self.evt_close.connect(lambda: hp.disable_widgets(*widgets, disabled=False))
        hp.disable_widgets(*widgets, disabled=True)

    def closeEvent(self, event: QCloseEvent | None = None) -> None:
        """Hide rather than close."""
        if self.HIDE_WHEN_CLOSE:
            self.hide()
            self.clearFocus()
            if hasattr(self, "evt_hide"):
                self.evt_hide.emit()
            event.ignore()
        else:
            self._on_teardown()
            self.connect_events(False)
            if hasattr(self, "evt_close"):
                self.evt_close.emit()
            super().closeEvent(event)

    def close(self) -> bool:
        """Hide dialog rather than delete it."""
        if self.HIDE_WHEN_CLOSE:
            self.hide()
            if hasattr(self, "evt_hide"):
                self.evt_hide.emit()
            self.clearFocus()
            return False
        else:
            return super().close()


class QtFramelessTool(QtFramelessPopup):
    """Frameless dialog that stays on top."""

    def __init__(
        self,
        parent: QWidget | None,
        title: str = "",
        position: QPoint | None = None,
        flags: Qt.WindowType = Qt.WindowType.FramelessWindowHint
        | Qt.WindowType.WindowStaysOnTopHint
        | Qt.WindowType.Tool,
        delay: bool = False,
    ):
        super().__init__(parent, title, position, flags, delay=delay)


class QtCollapsibleFramelessTool(QtFramelessTool):
    """Collapsible tool."""

    GEOM_TIME = 250

    expand_btn: QtExpandButton

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.timer = QTimer()
        self.geom_anim = QPropertyAnimation(self, b"geometry", self)

    def mousePressEvent(self, event):
        """Mouse press event."""
        event.ignore()

    def closeEvent(self, event: QCloseEvent | None) -> None:
        """Cannot close popup."""
        if event:
            event.ignore()

    @property
    def is_expanded(self) -> bool:
        """Checks whether text is expanded."""
        return bool(self.property("expanded"))

    def toggle_expansion(self):
        """Toggle the expanded state of the notification frame."""
        self.contract() if self.is_expanded else self.expand()
        self.timer.stop()

    def expand(self) -> None:
        """Expanded widget to maximum size."""
        sz = self.parent().size() - QSize(self.size().width(), self.parent().size().height())
        self.geom_anim.setDuration(self.GEOM_TIME)
        self.geom_anim.setStartValue(self.geometry())
        size = self.maximumSize()
        self.geom_anim.setEndValue(
            QRect(
                sz.width() - size.width() + self.minimumSize().width() - 20,
                sz.height() + 20,
                size.width(),
                size.height(),
            )
        )
        self.geom_anim.setEasingCurve(QEasingCurve.Type.OutQuad)
        self.geom_anim.start()
        self.expand_btn.expanded = True
        self.setProperty("expanded", True)
        self._widget.show()

    def contract(self):
        """Contract widget to minimum size."""
        sz = self.parent().size() - QSize(self.size().width(), self.parent().size().height())
        self.geom_anim.setDuration(self.GEOM_TIME)
        self.geom_anim.setStartValue(self.geometry())
        size = self.minimumSize()
        self.geom_anim.setEndValue(
            QRect(
                sz.width() + self.maximumSize().width() - size.width() - 20,
                sz.height() + 20,
                size.width(),
                size.height(),
            )
        )
        self.geom_anim.setEasingCurve(QEasingCurve.OutQuad)
        self.geom_anim.start()
        self.expand_btn.expanded = False
        self.setProperty("expanded", False)
        hp.polish_widget(self.expand_btn)
        self._widget.hide()

    def move_to_top_right(self, offset=(-20, 20)):
        """Position widget at the top right edge of the parent."""
        if not self.parent():
            return
        sz = self.parent().size() - QSize(self.size().width(), self.parent().size().height()) + QSize(*offset)
        self.move(QPoint(sz.width(), sz.height()))

    def move_to_bottom_right(self, offset=(20, 20)):
        """Position widget at the top right edge of the parent."""
        if not self.parent():
            return
        sz = self.parent().size() - self.size() - QSize(*offset)
        # sz = self.parent().size() - QSize(self.size().width(), self.parent().size().height()) + QSize(*offset)
        self.move(QPoint(sz.width(), sz.height()))


class QtTransparentPopup(QDialog, DialogMixin):
    """A generic popup window.

    The seemingly extra frame here is to allow rounded corners on a truly
    transparent background.  New items should be added to QtPopup.frame

    +----------------------------------
    | Dialog
    |  +-------------------------------
    |  | QVBoxLayout
    |  |  +----------------------------
    |  |  | QFrame
    |  |  |  +-------------------------
    |  |  |  |
    |  |  |  |  (add a new layout here)

    Parameters
    ----------
    parent : qtpy.QtWidgets:QWidget
        Parent widget of the popup dialog box.

    Attributes
    ----------
    frame : qtpy.QtWidgets.QFrame
        Frame of the popup dialog box.
    layout : qtpy.QtWidgets.QVBoxLayout
        Layout of the popup dialog box.
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.setObjectName("QtModalPopup")
        self.setModal(False)  # if False, then clicking anywhere else closes it
        self.setWindowFlags(Qt.Popup | Qt.FramelessWindowHint)
        self.setLayout(QVBoxLayout())

        self.frame = QFrame()
        self.frame.setObjectName("QtPopupFrame")
        self.layout().addWidget(self.frame)
        self.layout().setContentsMargins(0, 0, 0, 0)

        EVENTS.evt_force_exit.connect(self.close)

    def keyPressEvent(self, event):
        """Close window on return, else pass event through to super class.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            return self.close()
        super().keyPressEvent(event)


class SubWindowBase(QDialog):
    """Sub-window mixin."""

    # Animation attributes
    FADE_IN_RATE = 220
    FADE_OUT_RATE = 120
    MAX_OPACITY = 0.9
    # Window attributes
    MIN_WIDTH = 250
    MAX_WIDTH = 350
    MIN_HEIGHT = 40

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setWindowFlags(Qt.WindowType.SubWindow)
        self.setSizeGripEnabled(False)
        self.setModal(False)
        self.setMouseTracking(True)

        self.setMinimumWidth(self.MIN_WIDTH)
        self.setMaximumWidth(self.MAX_WIDTH)
        self.setMinimumHeight(self.MIN_HEIGHT)

        # opacity effect
        self.opacity = QGraphicsOpacityEffect()
        self.setGraphicsEffect(self.opacity)
        self.opacity_anim = QPropertyAnimation(self.opacity, b"opacity", self)
        # geometry effect
        self.geom_anim = QPropertyAnimation(self, b"geometry", self)

    def move_to(self, location: str) -> None:
        """Move to location."""
        try:
            if location == "top_right":
                self.move_to_top_right()
            elif location == "top_left":
                self.move_to_top_left()
            elif location == "bottom_right":
                self.move_to_bottom_right()
            elif location == "bottom_left":
                self.move_to_bottom_left()
        except RuntimeError:
            pass

    def move_to_top_right(self, offset=(-10, 10)) -> None:
        """Position widget at the top right edge of the parent."""
        if not self.parent():
            return
        psz = self.parent().size()
        sz = psz - QSize(self.size().width(), psz.height()) + QSize(*offset)
        self.move(QPoint(sz.width(), sz.height()))

    def move_to_bottom_right(self, offset=(8, 8)) -> None:
        """Position widget at the bottom right edge of the parent."""
        if not self.parent():
            return
        sz = self.parent().size() - self.size() - QSize(*offset)
        self.move(QPoint(sz.width(), sz.height()))

    def move_to_top_left(self, offset=(8, 8)) -> None:
        """Position widget at the bottom right edge of the parent."""
        if not self.parent():
            return
        self.move(QPoint(*offset))

    def move_to_bottom_left(self, offset=(8, 8)) -> None:
        """Position widget at the bottom right edge of the parent."""
        if not self.parent():
            return
        sz = self.parent().size() - self.size()
        self.move(QPoint(offset[0], sz.height() - offset[1]))

    def slide_in(self) -> None:
        """Run animation that fades in the dialog with a slight slide up."""
        geom = self.geometry()
        self.geom_anim.setDuration(self.FADE_IN_RATE)
        self.geom_anim.setStartValue(geom.translated(0, -20))
        self.geom_anim.setEndValue(geom)
        self.geom_anim.setEasingCurve(QEasingCurve.Type.OutQuad)
        # fade in
        self.opacity_anim.setDuration(self.FADE_IN_RATE)
        self.opacity_anim.setStartValue(0)
        self.opacity_anim.setEndValue(self.MAX_OPACITY)
        self.geom_anim.start()
        self.opacity_anim.start()

    def fade_in(self) -> None:
        """Run animation that fades in the dialog."""
        self.opacity_anim.setDuration(self.FADE_IN_RATE)
        self.opacity_anim.setStartValue(0)
        self.opacity_anim.setEndValue(self.MAX_OPACITY)
        self.opacity_anim.start()

    def fade_out(self) -> None:
        """Run animation that fades out the dialog."""
        self.opacity_anim.setDuration(self.FADE_OUT_RATE)
        self.opacity_anim.setStartValue(self.MAX_OPACITY)
        self.opacity_anim.setEndValue(0)
        self.opacity_anim.start()

    def deleteLater(self) -> None:
        """Stop all animations and timers before deleting."""
        with suppress(RuntimeError, TypeError):
            self.geom_anim.stop()
            super().deleteLater()

    def close(self) -> None:
        """Close window."""
        with suppress(RuntimeError, TypeError):
            self.geom_anim.stop()
            super().close()

    def close_with_fade(self):
        """Fade out then close."""
        with suppress(RuntimeError, TypeError):
            self.opacity_anim.stop()
            self.geom_anim.stop()

            self.opacity_anim.setDuration(self.FADE_OUT_RATE)
            self.opacity_anim.setStartValue(self.MAX_OPACITY)
            self.opacity_anim.setEndValue(0)
            self.opacity_anim.start()
            self.opacity_anim.finished.connect(self.close)



"""Task widget."""

from __future__ import annotations

import os
import typing as ty

from koyo.timer import MeasureTimer
from loguru import logger
from qtpy.QtCore import Qt, QTimer, Signal  # type: ignore[attr-defined]
from qtpy.QtWidgets import QFrame, QGridLayout, QWidget

import qtextra.helpers as hp
from qtextra.queue.task import Task
from qtextra.queue.utilities import format_command, format_interval
from qtextra.typing import TaskState
from qtextra.widgets.qt_button_icon import QtPauseButton

if ty.TYPE_CHECKING:
    from qtextra.queue.info import TaskInfoDialog

logger = logger.bind(src="TaskWidget")

IS_DEV = os.environ.get("DEV_MODE", "0") == "1"


class TaskWidget(QFrame):
    """Widget controlling and displaying task information."""

    evt_start_task = Signal(Task)
    evt_requeue_task = Signal(Task)
    evt_cancel_task = Signal(Task)
    evt_pause_task = Signal(Task, bool)
    evt_console = Signal(object)

    task: Task | None
    can_cancel: bool
    can_pause: bool
    can_cancel_when_started: bool
    can_force_start: bool

    dlg_info: TaskInfoDialog | None = None

    def __init__(self, parent: QWidget | None = None, toggled: bool = True):
        super().__init__(parent)
        self.setFrameShape(QFrame.Shape.Box)
        self.setLineWidth(1)

        self.poll_timer = QTimer(self)
        self.poll_timer.setInterval(1000)
        self.poll_timer.timeout.connect(self.on_update_timer)

        self.task_name = hp.make_label(
            self,
            "",
            alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
            enable_url=True,
            click_func=self.on_toggle_visibility,
            elide_mode=Qt.TextElideMode.ElideRight,
        )
        self.task_state = hp.make_label(
            self,
            "",
            alignment=Qt.AlignmentFlag.AlignCenter,
            object_name="task_state",
        )
        self.task_state.setMaximumWidth(90)

        self.task_info = hp.make_label(
            self,
            "",
            alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
            enable_url=True,
        )
        self.time_info = hp.make_label(self, "", alignment=Qt.AlignmentFlag.AlignCenter)

        self.options_btn = hp.make_qta_btn(
            self,
            "settings",
            tooltip="Show extra actions.",
            normal=True,
            func=self.on_open_menu,
        )
        self.options_btn.hide()  # disable for now

        self.clipboard_btn = hp.make_qta_btn(
            self,
            "clipboard",
            tooltip="Copy CLI commands to the clipboard.",
            normal=True,
            func=self.on_copy_to_clipboard,
        )
        self.info_btn = hp.make_qta_btn(
            self,
            "info",
            tooltip="Show information about the task.",
            normal=True,
            func=self.on_task_info,
        )

        self.start_btn = hp.make_qta_btn(
            self,
            "run",
            tooltip="Start task if the task has not started yet. This will override any built-in restrictions on number"
            " of simultaneous tasks and can cause your system to freeze.",
            normal=True,
            func=self.on_start_task,
        )
        self.retry_btn = hp.make_qta_btn(
            self, "retry", tooltip="Retry running task if the task has failed.", normal=True, func=self._on_retry_task
        )
        self.pause_btn = QtPauseButton(self)
        self.pause_btn.set_normal()
        self.pause_btn.setToolTip("Pause running task.")
        self.pause_btn.clicked.connect(self._on_pause_task)  # type: ignore[unused-ignore]
        self.cancel_btn = hp.make_qta_btn(
            self,
            "cross_full",
            tooltip="Cancel task. If a task has started, this is not guaranteed to work!",
            normal=True,
            func=self._on_cancel_task,
        )
        self.task_id = hp.make_label(
            self,
            "",
            alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
            object_name="task_id",
        )

        layout = QGridLayout(self)
        # widget, row, column, rowspan, colspan
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        layout.setColumnStretch(0, 1)
        # row 0
        layout.addWidget(self.task_name, 0, 0, 1, 1)
        layout.addWidget(self.time_info, 0, 1, 1, 1)
        layout.addWidget(self.task_state, 0, 2, 1, 1)
        # row 1
        layout.addWidget(self.task_info, 1, 0, 1, 3)
        # row 2
        layout.addLayout(
            hp.make_h_layout(
                self.options_btn,
                self.clipboard_btn,
                self.info_btn,
                self.start_btn,
                self.retry_btn,
                self.pause_btn,
                self.cancel_btn,
                stretch_before=True,
                spacing=2,
            ),
            1,
            1,
            1,
            2,
        )
        # row 3
        layout.addWidget(self.task_id, 3, 0, 1, 1)

        # setup buttons
        hp.disable_widgets(self.retry_btn, self.pause_btn, disabled=True)
        self.toggled = toggled

    @property
    def toggled(self) -> bool:
        """Return state of toggled."""
        return self._toggled

    @toggled.setter
    def toggled(self, value: bool) -> None:
        self._toggled = value
        self.toggle_visibility()

    def on_toggle_visibility(self) -> None:
        """Toggle visibility."""
        self.toggled = not self.toggled

    def toggle_visibility(self) -> None:
        """Hide certain widgets."""
        hp.hide_widgets(
            self.task_info,
            self.start_btn,
            self.retry_btn,
            self.pause_btn,
            self.cancel_btn,
            # self.options_btn,
            self.info_btn,
            self.clipboard_btn,
            hidden=self.toggled,
        )

    def set_task(
        self,
        task: Task,
        can_cancel: bool = False,
        can_pause: bool = False,
        can_cancel_when_started: bool = False,
        can_force_start: bool = False,
        auto_expand: bool = True,
    ) -> None:
        """Setup UI for task."""
        with MeasureTimer() as timer:
            self.task = task
            self.can_cancel = can_cancel
            self.can_cancel_when_started = can_cancel_when_started
            self.can_pause = can_pause
            self.can_force_start = can_force_start

            # update ui
            self.task_name.setText(task.task_name_repr or hp.hyper(task.task_name))
            self.task_name.setToolTip(task.task_name_tooltip or task.task_name)
            self.task_info.setText(task.pretty_info)
            self.task_id.setText(f"Task ID: {task.task_id}")
            if task.state == TaskState.FINISHED:
                self.stop()
            else:
                hp.disable_widgets(self.start_btn, self.cancel_btn, disabled=False)
            self._update_state()
            self._update_warnings()
            self.toggled = not auto_expand
        logger.trace(f"Added task '{self.task.summary()}' in {timer()}")

    def _update_warnings(self) -> None:
        """Generate warnings for task."""
        # errors: list[str] = []
        # if self.task:
        #     errors = self.task.config_errors()
        # tooltip = "<br>".join(errors)
        # state, color = get_icon_state(errors)
        # self.errors_btn.setIcon(hp.make_qta_icon(state, color=color))  # type: ignore[no-untyped-call]
        # self.errors_btn.setToolTip(tooltip)

    def on_copy_to_clipboard(self) -> None:
        """Copy commands to clipboard."""
        if self.task:
            commands = list(self.task.command_iter())
            commands = [" ".join(cmd) for cmd in commands]
            hp.copy_text_to_clipboard(format_command(commands, IS_DEV))
            hp.add_flash_animation(self, duration=1000)

    def on_task_info(self) -> None:
        """Show widget with information about the task."""
        if self.task:
            from qtextra.queue.info import TaskInfoDialog

            try:
                if self.dlg_info is None:
                    self.dlg_info = TaskInfoDialog(self, self.task)
                    self.dlg_info.evt_update.connect(self.on_update_timer)
                self.dlg_info.show()
                self.dlg_info.raise_()
            except RuntimeError:
                self.dlg_info = None
                logger.error("Failed to show task info dialog.")
                self.on_task_info()

    def update_progress(self) -> None:
        """Update progress."""
        try:
            if self.dlg_info:
                self.dlg_info.update_progress()
        except (AttributeError, RuntimeError, Exception):
            self.dlg_info = None

    def on_open_menu(self) -> None:
        """Open folder menu."""

    def _on_retry_task(self) -> None:
        """Try task again."""
        task = self.task
        if task:
            task.state = TaskState.QUEUED
            self.evt_requeue_task.emit(task)  # type: ignore[unused-ignore]

    def on_start_task(self) -> None:
        """Triggered when user clicked on the run task button."""
        if self.task:
            self.evt_start_task.emit(self.task)  # type: ignore[unused-ignore]
            self.started()

    def _update_state(self) -> None:
        """Update state."""
        if self.task:
            self.task_info.setText(self.task.pretty_info)
            self.task_state.setText(self.task.state.capitalize())
            hp.polish_widget(self.task_state)
            self.update_progress()
            logger.trace(f"Updating task '{self.task.summary()}'...")

    def started(self) -> None:
        """Start task."""
        self.poll_timer.start()
        hp.disable_widgets(self.pause_btn, disabled=self.can_pause)
        hp.disable_widgets(self.start_btn, disabled=True)
        self._update_state()

    def _on_pause_task(self) -> None:
        """Triggered when user clicked to pause task."""
        if self.task:
            self.pause_btn.paused = not self.pause_btn.paused
            self.evt_pause_task.emit(self.task, self.pause_btn.paused)  # type: ignore[unused-ignore]
        self._update_state()

    def paused(self, paused: bool) -> None:
        """Task was paused."""
        if self.task:
            self.pause_btn.paused = paused
            if self.pause_btn.paused:
                self.poll_timer.stop()
                logger.trace(f"Pausing task '{self.task.summary()}'...")
            else:
                self.poll_timer.start()
                logger.trace(f"Restarting task '{self.task.summary()}'...")
            self._update_state()

    def _on_cancel_task(self, force: bool = False) -> None:
        """Triggered when user clicked to pause task."""
        if self.task:
            if force or hp.confirm(self, "Are you sure you wish to cancel this task?", "Cancel task?"):
                self.evt_cancel_task.emit(self.task)  # type: ignore[unused-ignore]
        self._update_state()

    def cancel(self) -> None:
        """Triggered when user clicked to cancel task."""
        self._on_cancel_task()

    def cancelled(self) -> None:
        """Task was cancelled."""
        if self.task:
            try:
                self.poll_timer.stop()
            except RuntimeError:
                return
            hp.disable_widgets(self.start_btn, self.pause_btn, self.cancel_btn, disabled=True)
            hp.disable_widgets(self.retry_btn, disabled=False)
            duration = self.task.duration
            self.time_info.setText(format_interval(duration))
            logger.trace(f"Cancelled task '{self.task.summary()}'...")
        self._update_state()

    def next(self) -> None:
        """Update task."""
        self._update_state()

    def part_failed(self) -> None:
        """Update task."""
        self._update_state()

    def stop(self) -> None:
        """Stop task."""
        try:
            self.poll_timer.stop()
        except RuntimeError:
            return
        if self.task:
            self.time_info.setText(format_interval(self.task.duration))
        hp.disable_widgets(self.start_btn, self.pause_btn, self.cancel_btn, disabled=True)
        hp.disable_widgets(self.retry_btn, disabled=False)
        self._update_state()

    def close(self) -> bool:
        """Close method."""
        self.poll_timer.timeout.disconnect(self.on_update_timer)
        self.poll_timer.stop()
        return super().close()

    def on_update_timer(self) -> None:
        """Update stats about the process."""
        if self.task:
            self.time_info.setText(format_interval(self.task.current_duration))

    def mousePressEvent(self, event: ty.Any) -> None:
        """Mouse press event."""
        self.toggled = not self.toggled
        return QFrame.mousePressEvent(self, event)



"""Test CLI Queue."""

import pytest
from koyo.system import IS_WIN

from qtextra.queue.cli_queue import CLIQueueHandler
from qtextra.queue.task import Task


@pytest.fixture
def setup_widget():
    """Setup panel"""

    def _widget() -> CLIQueueHandler:
        widget = CLIQueueHandler()
        return widget

    return _widget


class TestCLIQueueHandler:
    @pytest.mark.xfail(IS_WIN, reason="Some signals don't fire on Windows in time")
    def test_init(self, qtbot, setup_widget):
        queued, started, next_, finished, errored, cancelled, paused = [], [], [], [], [], [], []  # type: ignore
        queue = setup_widget()
        queue.n_parallel = 10
        queue.evt_queued.connect(queued.append)
        queue.evt_started.connect(started.append)
        queue.evt_next.connect(next_.append)
        queue.evt_finished.connect(finished.append)

        def _append_error(task, other=None) -> None:
            errored.append(task)

        queue.evt_errored.connect(_append_error)
        queue.evt_cancelled.connect(cancelled.append)

        def _append_pause(task, other=None) -> None:
            paused.append(task)

        queue.evt_paused.connect(_append_pause)
        assert queue, "Queue is not initialized"
        assert queue.n_tasks == (0, 0), "Queue should be empty"

        # test addition/running
        with (
            qtbot.assertNotEmitted(queue.evt_started),
            qtbot.assertNotEmitted(queue.evt_finished),
            qtbot.assertNotEmitted(queue.evt_errored),
        ):
            task = Task(task_id="123", task_name="Task 1", commands=[["echo", "Hello World"]])
            task_id = queue.add_task(task)
            assert len(queued) == 1, "Queue should have one task"
            assert queue.n_tasks == (0, 1), "Queue should have one task"
            assert task_id == task.task_id, "Task ID should be the same"
            assert not queue.queue_closed, "Queue should not be closed"
            assert not queue.is_running(), "Queue should not be running"
            assert len(queue.pending_queue) == 1, "Queue should have one pending task"
            assert len(queue.running_queue) == 0, "Queue should not have any running tasks"

        with qtbot.waitSignals([queue.evt_started, queue.evt_finished], timeout=1000):
            queue.run_queued()

        assert len(started) == 1, "Queue should have one task running"
        assert len(errored) == 0, "Queue should not have any errors"
        assert len(finished) == 1, "Queue should have one task finished"
        assert queue.n_tasks == (0, 0), "Queue should have one task running"
        assert len(queue.pending_queue) == 0, "Queue should have one pending task"
        assert len(queue.running_queue) == 0, "Queue should not have any running tasks"

        # test failure
        with (
            qtbot.assertNotEmitted(queue.evt_started),
            qtbot.assertNotEmitted(queue.evt_finished),
            qtbot.assertNotEmitted(queue.evt_errored),
        ):
            task = Task(
                task_id="234",
                task_name="Task 2",
                commands=[
                    ["sleep", "0.5"],
                    ["fake-command", "Hello World"],  # should fail
                ],
            )  # error in command
            task_id = queue.add_task(task)
            assert task_id == task.task_id, "Task ID should be the same"
            assert len(queued) == 2, "Queue should have one task"

        with qtbot.waitSignals([queue.evt_started, queue.evt_errored], timeout=1000):
            queue.run_queued()
        assert len(started) == 2, "Queue should have one task running"
        assert len(errored) == 1, "Queue should not have one error"
        assert len(finished) == 1, "Queue should have one task finished"

        # test cancellation
        with (
            qtbot.assertNotEmitted(queue.evt_started),
            qtbot.assertNotEmitted(queue.evt_finished),
            qtbot.assertNotEmitted(queue.evt_errored),
        ):
            task = Task(
                task_id="345", task_name="Task 3", commands=[["sleep", "0.5"], ["echo", "Hello World"]]
            )  # error in command
            task_id = queue.add_task(task)
            assert task_id == task.task_id, "Task ID should be the same"
            assert len(queued) == 3, "Queue should have one task"

        with qtbot.waitSignals([queue.evt_started], timeout=500):
            queue.run_queued()
        with qtbot.waitSignals([queue.evt_cancelled], timeout=5000 if IS_WIN else 1500):
            queue.cancel(task)

        assert len(started) == 3, "Queue should have one task running"
        assert len(cancelled) == 1, "Queue should have one task finished"

        # test pause
        with (
            qtbot.assertNotEmitted(queue.evt_started),
            qtbot.assertNotEmitted(queue.evt_finished),
            qtbot.assertNotEmitted(queue.evt_errored),
        ):
            task = Task(
                task_id="456", task_name="Task 4", commands=[["sleep", "0.5"], ["echo", "Hello World"]]
            )  # error in command
            task_id = queue.add_task(task)
            assert task_id == task.task_id, "Task ID should be the same"
            assert len(queued) == 4, "Queue should have one task"

        assert queue.n_parallel > 0, "Queue should have some parallel tasks"
        assert queue.is_available(), "Queue should be available"
        assert len(queue.pending_queue) > 0, "Queue should have some tasks"
        queue.auto_run = True
        with qtbot.waitSignals([queue.evt_started], timeout=500):
            queue.run_queued()
        with qtbot.waitSignals([queue.evt_paused], timeout=1500):
            queue.pause(task, True)
        assert len(started) == 4, "Queue should have one task running"
        assert len(paused) == 1, "Queue should have one task finished"

        with qtbot.waitSignals([queue.evt_started], timeout=1500):
            queue.pause(task, False)
        assert len(paused) == 1, "Queue should have one task finished"



"""Themes configuration file."""

import re
import typing as ty
from functools import lru_cache
from itertools import product
from pathlib import Path

import numpy as np
from koyo.timer import MeasureTimer
from loguru import logger
from psygnal import EventedModel
from pydantic import ConfigDict, PrivateAttr, ValidationError, field_validator
from pydantic_extra_types.color import Color
from qtpy.QtCore import QDateTime, QTime, Signal
from qtpy.QtGui import QColor, QPalette
from qtpy.QtWidgets import QApplication, QWidget

from qtextra.config.config import ConfigBase, _get_previous_configs
from qtextra.utils.appdirs import USER_CACHE_DIR

DARK_THEME = {
    "name": "dark",
    "type": "dark",
    "background": "rgb(38, 41, 48)",
    "foreground": "rgb(65, 72, 81)",
    "primary": "rgb(90, 98, 108)",
    "secondary": "rgb(189, 147, 249)",
    "highlight": "rgb(106, 115, 128)",
    "text": "rgb(240, 241, 242)",
    "icon": "rgb(209, 210, 212)",
    "warning": "rgb(255, 105, 60)",
    "error": "rgb(183, 52, 53)",
    "success": "rgb(30, 215, 96)",
    "progress": "rgb(179, 98, 0)",
    "current": "rgb(0, 122, 204)",
    "syntax_style": "native",
    "console": "rgb(0, 0, 0)",
    "canvas": "rgb(0, 0, 0)",
    "standout": "rgb(255, 255, 0)",
    "font_size": "14pt",
    "header_size": "18pt",
}
LIGHT_THEME = {
    "name": "light",
    "type": "light",
    "background": "rgb(239, 235, 233)",
    "foreground": "rgb(214, 208, 206)",
    "primary": "rgb(188, 184, 181)",
    "secondary": "rgb(190, 185, 183)",
    "highlight": "rgb(163, 158, 156)",
    "text": "rgb(59, 58, 57)",
    "icon": "rgb(107, 105, 103)",
    "warning": "rgb(255, 105, 60)",
    "error": "rgb(255, 18, 31)",
    "success": "rgb(30, 215, 96)",
    "progress": "rgb(255, 175, 77)",
    "current": "rgb(30, 215, 96)",
    "syntax_style": "default",
    "console": "rgb(255, 255, 255)",
    "canvas": "rgb(255, 255, 255)",
    "standout": "rgb(255, 252, 0)",
    "font_size": "14pt",
    "header_size": "18pt",
}


def time_to_qt_time(value: str) -> QTime:
    """Parse config time to QTime format."""
    time = QTime()
    try:
        _, _ = value.split(":")
    except Exception:
        return time
    time = time.fromString(value, "HH:mm")
    return time


def parse_time(value: str) -> ty.Tuple[int, ...]:
    """Parse time."""
    try:
        hh, mm = value.split(":")
        return int(hh), int(mm)
    except Exception:
        return -1, -1


class CanvasTheme(EventedModel):
    """Plot theme model."""

    canvas: Color
    line: Color
    scatter: Color
    highlight: Color
    axis: Color
    gridlines: Color
    label: Color
    _canvas_backup: ty.Optional[Color] = PrivateAttr(None)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._canvas_backup = self.canvas

    def as_array(self, name: str) -> np.ndarray:
        """Return color array."""
        return np.asarray(getattr(self, name))


class CanvasThemes(ConfigBase):
    """Plot theme class."""

    # event emitted whenever a theme is changed
    evt_theme_changed = Signal()

    def __init__(self):
        super().__init__(None)
        self.themes = {}
        self._theme = "light"
        self._integrate_canvas: bool = False

        self.add_theme(
            "dark",
            CanvasTheme(
                canvas="black",
                line="white",
                scatter="white",
                highlight="yellow",
                axis="white",
                gridlines="white",
                label="lightgray",
            ),
        )
        self.add_theme(
            "light",
            CanvasTheme(
                canvas="white",
                line="black",
                scatter="black",
                highlight="yellow",
                axis="black",
                gridlines="black",
                label="black",
            ),
        )

        for theme in self.themes.values():
            theme.events.connect(lambda _: self.evt_theme_changed.emit())

    @property
    def integrate_canvas(self):
        """Integrate canvas with background color."""
        return self._integrate_canvas

    @integrate_canvas.setter
    def integrate_canvas(self, value):
        self._integrate_canvas = value
        background = THEMES.active.background if value else self.active._canvas_backup
        self.active.canvas = background

    def add_theme(self, name: str, theme_data: ty.Union[CanvasTheme, ty.Dict[str, str]]):
        """Add theme."""
        if isinstance(theme_data, CanvasTheme):
            self.themes[name] = theme_data
        else:
            self.themes[name] = CanvasTheme(**theme_data)

    def available_themes(self) -> ty.Tuple[str, ...]:
        """Get list of available themes."""
        return tuple(self.themes)

    @property
    def active(self) -> CanvasTheme:
        """Return active theme."""
        return self.themes[self.theme]

    @property
    def theme(self) -> str:
        """Return theme name."""
        return self._theme

    @theme.setter
    def theme(self, value: str):
        if self._theme == value:
            return
        if value not in self.themes:
            return
        self._theme = value
        self.integrate_canvas = self._integrate_canvas
        self.evt_theme_changed.emit()

    # def as_array(self, name: str) -> np.ndarray:
    #     """Return color array."""
    #     from napari.utils.colormaps.standardize_color import transform_color
    #
    #     color: Color = getattr(self.active, name)
    #     return transform_color(color.as_hex())[0]

    def as_hex(self, name: str) -> str:
        """Return color as hex."""
        color: Color = getattr(self.active, name)
        return color.as_hex()


class Theme(EventedModel):
    """Theme model.

    Attributes
    ----------
    name : str
        Name of the virtual folder where icons will be saved to.
    syntax_style : str
        Name of the console style.
        See for more details: https://pygments.org/docs/styles/
    canvas : Color
        Background color of the canvas.
    background : Color
        Color of the application background.
    foreground : Color
        Color to contrast with the background.
    primary : Color
        Color used to make part of a widget more visible.
    secondary : Color
        Alternative color used to make part of a widget more visible.
    highlight : Color
        Color used to highlight visual element.
    text : Color
        Color used to display text.
    warning : Color
        Color used to indicate something is wrong.
    current : Color
        Color used to highlight Qt widget.
    """

    model_config = ConfigDict(validate_assignment=True)

    name: str
    type: str
    syntax_style: str
    canvas: Color
    console: Color
    background: Color
    foreground: Color
    primary: Color
    secondary: Color
    highlight: Color
    text: Color
    icon: Color
    warning: Color
    error: Color
    success: Color
    current: Color
    progress: Color
    standout: Color
    font_size: str = "14pt"
    header_size: str = "18pt"

    def __getattr__(self, item: str) -> ty.Union[str, Color]:
        return getattr(self, item)

    @field_validator(
        "canvas",
        "console",
        "background",
        "foreground",
        "primary",
        "secondary",
        "highlight",
        "text",
        "icon",
        "warning",
        "error",
        "success",
        "current",
        "progress",
        "standout",
        mode="before",
    )
    def _validate_color(value) -> Color:
        if isinstance(value, np.ndarray):
            value = value.tolist()
        elif isinstance(value, str):
            value = Color(value).as_hex()
        return Color(value)

    @field_validator("syntax_style", mode="before")
    def _ensure_syntax_style(value: str) -> str:
        from pygments.styles import STYLE_MAP

        assert value in STYLE_MAP, (
            f"Incorrect `syntax_style` value provided. Please use one of the following: {', '.join(STYLE_MAP)}"
        )
        return value

    @field_validator("font_size", "header_size", mode="before")
    def _ensure_font_size(value: ty.Union[int, str]) -> str:
        if isinstance(value, int):
            value = str(value)
        value = value.replace("px", "").replace("pt", "")
        if not value.endswith("pt"):
            return value + "pt"
        return value

    @property
    def id(self) -> str:
        """Return theme id."""
        return self.name

    def to_dict(self) -> dict:
        """Export as dictionary."""
        data = {}
        for key, value in self:
            if isinstance(value, Color):
                data[key] = value.as_hex()
            else:
                data[key] = value
        return data


class Themes(ConfigBase):
    """Themes."""

    DEFAULT_CONFIG_NAME = "themes-config.json"
    DEFAULT_CONFIG_GROUPS = ("settings",)
    EXTRA_CONFIG_GROUPS = ("themes",)

    REQUIRED_KEYS = (
        "name",  # name of the theme so the icons can be placed there
        "type",  # type of theme - either dark or light
        "background",  # background color
        "foreground",  # foreground color
        "primary",  # primary color for highlights
        "secondary",  # secondary color for less intense highlights
        "highlight",  # highlight color
        "text",  # color of text
        "icon",  # color of icons
        "warning",  # color of warning
        "error",  # color of error
        "success",  # color of success
        "progress",  # color of progress
        "current",  # color of current item
        "syntax_style",  # used by console
        "console",  # used by console
        "canvas",  # color of the canvas
        "standout",  # standout color
    )

    # event emitted whenever a new theme is added
    evt_theme_added = Signal()
    # event emitted whenever a theme is changed
    evt_theme_changed = Signal()
    # event emitted whenever user changed time check
    evt_update_timer = Signal()
    # event emitted whenever icon color is changed
    evt_theme_icon_changed = Signal()
    # event emitted whenever stylesheets are changed
    evt_qss_changed = Signal()

    def __init__(self):
        super().__init__(None)

        self._theme: str = "light"
        self._sync_with_time: bool = True
        self._light_start_time: str = "08:00"
        self._light_end_time: str = "20:00"
        self.themes: ty.Dict[str, Theme] = {}
        self.add_theme(
            "dark",
            Theme(**DARK_THEME),
        )
        self.add_theme(
            "light",
            Theme(**LIGHT_THEME),
        )
        # synchronize our icon with napari icon color
        try:
            from napari.utils.theme import _themes

            for name in _themes:
                _themes[name].icon = self.get_hex_color("icon")
        except ImportError:
            pass

        for theme in self.themes.values():
            theme.events.connect(lambda _: self.evt_theme_changed.emit())

    def __getitem__(self, item):
        return self.themes[item]

    @property
    def active(self) -> Theme:
        """Return active theme."""
        return self.themes[self.theme]

    def get_sync_theme(self) -> str:
        """Get theme based on synchronization settings."""
        curr_time: QTime = QDateTime.currentDateTime().time()
        start_time = time_to_qt_time(self.light_start_time)
        end_time = time_to_qt_time(self.light_end_time)
        if start_time <= curr_time <= end_time:
            return "light"
        return "dark"

    def synchronize_theme(self) -> None:
        """Synchronize theme."""
        if self.sync_with_time:
            theme = self.get_sync_theme()
            self.theme = theme

    @property
    def sync_with_time(self) -> bool:
        """Flag to indicate whether theme should be synchronized with time."""
        return self._sync_with_time

    @sync_with_time.setter
    def sync_with_time(self, value: bool):
        self._sync_with_time = value
        self.evt_update_timer.emit()

    @property
    def light_start_time(self) -> str:
        """Get morning time."""
        return self._light_start_time

    @light_start_time.setter
    def light_start_time(self, value: str) -> None:
        self._light_start_time = value
        self.synchronize_theme()

    @property
    def light_end_time(self) -> str:
        """Get evening time."""
        return self._light_end_time

    @light_end_time.setter
    def light_end_time(self, value: str) -> None:
        self._light_end_time = value
        self.synchronize_theme()

    @property
    def theme(self) -> str:
        """Get theme."""
        return self._theme

    @theme.setter
    def theme(self, value: str) -> None:
        """Set theme."""
        if self._theme == value:
            return
        self._theme = value
        # synchronize our icon with napari icon color
        try:
            from napari.utils.theme import _themes

            for name in _themes:
                _themes[name].icon = self.get_hex_color("icon")
        except ImportError:
            pass
        with MeasureTimer() as timer:
            self.evt_theme_changed.emit()
            self.evt_theme_icon_changed.emit()
        logger.debug(f"Changed theme to '{value}' in {timer()}")

    @staticmethod
    def update_palette() -> None:
        """Get updated palette."""
        qapp = QApplication.instance()
        if qapp is None:
            return
        palette = qapp.palette()
        palette.setColor(QPalette.ColorRole.Link, QColor("#f56833"))
        palette.setColor(QPalette.ColorRole.LinkVisited, QColor("#f56833"))
        qapp.setPalette(palette)

    @property
    def syntax_style(self) -> str:
        """Get syntax style."""
        return self.active.syntax_style

    def get_font_size(self) -> int:
        """Get font size."""
        unit = "pt" if "pt" in self.active.font_size else "px"
        return int(self.active.font_size.split(unit)[0])

    def get_rgb_color(self, name: str) -> str:
        """Get color in the default style."""
        color: Color = getattr(self.active, name)
        return color.as_rgb()

    def get_hex_color(self, name: str) -> str:
        """Get color in hex format."""
        color: Color = getattr(self.active, name)
        return color.as_hex(format="long")

    def get_qt_color(self, name: str) -> QColor:
        """Get QColor."""
        color = self.get_hex_color(name)
        return QColor(color)

    @staticmethod
    def get_text_color_for_background(color: str) -> str:
        """Get text color for background."""
        from qtextra.utils.color import get_text_color, rgb_to_hex

        color = QColor(color)
        return rgb_to_hex(get_text_color(color).getRgb())

    def get_theme(
        self, theme_name: ty.Optional[str] = None, as_dict: bool = False
    ) -> ty.Union[Theme, ty.Dict[str, str]]:
        """Get a theme based on its name.

        Parameters
        ----------
        theme_name : str
            Name of requested theme.
        as_dict : bool
            Flag to return as dictionary.

        Returns
        -------
        theme: dict of str: str
            Theme mapping elements to colors. A copy is created
            so that manipulating this theme can be done without
            side effects.
        """
        if theme_name is None:
            theme_name = self.theme
        if theme_name in self.themes:
            theme = self.themes[theme_name]
            _theme = theme.copy()
            if as_dict:
                return _theme.model_dump()
            return _theme
        else:
            raise ValueError(f"Unrecognized theme {theme_name}. Available themes are {self.available_themes()}")

    @property
    def is_dark(self) -> bool:
        """Check if theme is dark."""
        return self.active.type == "dark"

    def add_theme(self, name: str, theme_data: ty.Union[Theme, ty.Dict[str, str]], register: bool = False):
        """Add theme."""
        if name not in self.themes:
            self.add_resource(name)
        if isinstance(theme_data, dict):
            theme_data = Theme(**theme_data)

        self.themes[name] = theme_data
        self.themes[name].events.icon.connect(lambda _: self._emit_icon_color_change(name))
        if register:
            self.register_themes([name])

    def add_resource(self, name: str) -> None:
        """Add resources to QDir."""
        from qtpy.QtCore import QDir

        QDir.addSearchPath(f"theme_{name}", str(self.get_theme_path(name)))
        logger.debug(f"Added '{name}' theme to resources path")

    def register_themes(self, names: ty.Optional[ty.List[str]] = None) -> None:
        """Register themes."""
        from qtextra.icons import build_theme_svgs

        if names is None:
            names = list(self.themes.keys())

        for name in names:
            build_theme_svgs(name)

    def available_themes(self) -> ty.Tuple[str, ...]:
        """Get list of available themes."""
        return tuple(self.themes)

    def get_theme_color(self, key: str = "text", theme_name: ty.Optional[str] = None) -> str:
        """Get text color appropriate for the theme."""
        if theme_name is None:
            theme_name = self.theme
        palette = self.themes[theme_name]
        return getattr(palette, key).as_hex()

    def get_theme_stylesheet(self, theme_name: ty.Optional[str] = None) -> str:
        """Get stylesheet."""
        from qtextra.assets import get_stylesheet
        from qtextra.utils.template import template

        if theme_name is None:
            theme_name = self.theme
        palette = self.themes[theme_name].model_dump()
        stylesheet = get_stylesheet()
        stylesheet = template(stylesheet, **palette)
        return stylesheet

    def set_theme_stylesheet(self, widget: QWidget, theme_name: ty.Optional[str] = None) -> None:
        """Set stylesheet on widget."""
        widget.setStyleSheet(self.get_theme_stylesheet(theme_name))

    def apply(self, widget: QWidget) -> None:
        """Apply theme on widget."""
        self.set_theme_stylesheet(widget)

    @staticmethod
    def get_theme_path(theme_name: str) -> Path:
        """Get path of directory for a given theme name."""
        from qtextra.utils.appdirs import USER_THEME_DIR

        return USER_THEME_DIR / theme_name

    def _get_config_parameters(self, config: ty.Dict) -> ty.Dict:
        """Get configuration parameters."""
        config["themes"] = {}
        for name, theme in self.themes.items():
            config["themes"][name] = theme.to_dict()
        config["settings"] = {
            "theme": self.theme,
            "sync_with_time": self._sync_with_time,
            "light_start_time": self._light_start_time,
            "light_end_time": self._light_end_time,
        }
        return config

    def _emit_icon_color_change(self, name: str) -> None:
        """Emit icon color change event."""
        self.register_themes([name])
        self.evt_theme_icon_changed.emit()
        logger.debug(f"Updating icon color for '{name}'...")

    def _set_config_parameters(self, config: ty.Dict) -> None:
        """Set extra configuration parameters."""
        for config_group_title in ("themes",):
            _config_group = config.get(config_group_title, {})
            for theme_name, theme in _config_group.items():
                try:
                    if theme_name in self.themes:
                        for key, value in theme.items():
                            try:
                                setattr(self.themes[theme_name], key, value)
                            except ValidationError as err:
                                logger.warning(
                                    f"Failed setting of {key} because it did not pass validation."
                                    f"\nFailed with error=`{err}`"
                                )
                    else:
                        theme = Theme(**theme)
                        theme.events.icon.connect(lambda _: self._emit_icon_color_change(theme_name))  # noqa: B023
                        self.themes[theme_name] = theme
                except ValidationError as err:
                    logger.warning(
                        f"Skipping {theme_name} theme because it did not pass validation.\nFailed with error=`{err}`"
                    )
                except Exception:
                    logger.warning("Could not load theme data.")

    def lighten(self, color: str, percentage: float = 10, as_hex: bool = False) -> str:
        """Lighted color."""
        from qtextra.utils.template import lighten

        color = lighten(color, percentage)
        if as_hex:
            color = Color(color).as_hex()
        return color

    def darken(self, color: str, percentage: float = 10, as_hex: bool = False) -> str:
        """Lighted color."""
        from qtextra.utils.template import darken

        color = darken(color, percentage)
        if as_hex:
            color = Color(color).as_hex()
        return color


def get_previous_configs(base_dir: ty.Optional[str] = None, filename: str = "themes-config.json") -> ty.Dict[str, str]:
    """Return dictionary of version : path of previous configuration files."""
    return _get_previous_configs(base_dir, filename)


svg_elem = re.compile(r"(<svg[^>]*>)")
svg_style = """<style type="text/css">
path {{fill: {0}; opacity: {1};}}
polygon {{fill: {0}; opacity: {1};}}
circle {{fill: {0}; opacity: {1};}}
rect {{fill: {0}; opacity: {1};}}
</style>"""


def get_theme_cache_dir(theme_name: str) -> Path:
    """Get theme cache directory."""
    return Path(USER_CACHE_DIR) / "_themes" / theme_name


@lru_cache
def get_raw_svg(path: str) -> str:
    """Get and cache SVG XML."""
    return Path(path).read_text()


@lru_cache
def get_colorized_svg(path_or_xml: ty.Union[str, Path], color: ty.Optional[str] = None, opacity=1) -> str:
    """Return a colorized version of the SVG XML at ``path``.

    Raises
    ------
    ValueError
        If the path exists but does not contain valid SVG data.
    """
    path_or_xml = str(path_or_xml)
    xml = path_or_xml if "</svg>" in path_or_xml else get_raw_svg(path_or_xml)
    if not color:
        return xml

    if not svg_elem.search(xml):
        raise ValueError(f"Could not detect svg tag in {path_or_xml!r}")
    # use regex to find the svg tag and insert css right after
    # (the '\\1' syntax includes the matched tag in the output)
    return svg_elem.sub(f"\\1{svg_style.format(color, opacity)}", xml)


def generate_colorized_svgs(
    svg_paths: ty.Iterable[ty.Union[str, Path]],
    colors: ty.Iterable[ty.Union[str, ty.Tuple[str, str]]],
    opacities: ty.Iterable[float] = (1.0,),
    theme_override: ty.Optional[ty.Dict[str, str]] = None,
) -> ty.Iterator[ty.Tuple[str, str]]:
    """Helper function to generate colorized SVGs.

    This is a generator that yields tuples of ``(alias, icon_xml)`` for every
    combination (Cartesian product) of `svg_path`, `color`, and `opacity`
    provided. It can be used as input to :func:`_temporary_qrc_file`.

    Parameters
    ----------
    svg_paths : Iterable[Union[str, Path]]
        An iterable of paths to svg files
    colors : Iterable[Union[str, Tuple[str, str]]]
        An iterable of colors.  Every icon will be generated in every color. If
        a `color` item is a string, it should be valid svg color style. Items
        may also be a 2-tuple of strings, in which case the first item should
        be an available theme name
        (:func:`~napari.utils.theme.available_themes`), and the second item
        should be a key in the theme (:func:`~napari.utils.theme.get_theme`),
    opacities : Iterable[float], optional
        An iterable of opacities to generate, by default (1.0,) Opacities less
        than one can be accessed in qss with the opacity as a percentage
        suffix, e.g.: ``my_svg_50.svg`` for opacity 0.5.
    theme_override : Optional[Dict[str, str]], optional
        When one of the `colors` is a theme ``(name, key)`` tuple,
        `theme_override` may be used to override the `key` for a specific icon
        name in `svg_paths`.  For example ``{'exclamation': 'warning'}``, would
        use the theme "warning" color for any icon named "exclamation.svg" by
        default `None`

    Yields
    ------
    (alias, xml) : Iterator[Tuple[str, str]]
        `alias` is the name that will used to access the icon in the Qt
        Resource system (such as QSS), and `xml` is the *raw* colorzied SVG
        text (as read from a file, perhaps pre-colored using one of the below
        functions).
    """
    # mapping of svg_stem to theme_key
    theme_override = theme_override or {}

    ALIAS_T = "{color}/{svg_stem}{opacity}.svg"

    for color, path, op in product(colors, svg_paths, opacities):
        clrkey = color
        svg_stem = Path(path).stem
        if isinstance(color, tuple):
            clrkey, theme_key = color
            theme_key = theme_override.get(svg_stem, theme_key)
            color = getattr(THEMES.get_theme(clrkey, False), theme_key).as_hex()
            # convert color to string to fit get_colorized_svg signature

        op_key = "" if op == 1 else f"_{op * 100:.0f}"
        alias = ALIAS_T.format(color=clrkey, svg_stem=svg_stem, opacity=op_key)
        yield alias, get_colorized_svg(path, color, op)


def write_colorized_svgs(
    dest: ty.Union[str, Path],
    svg_paths: ty.Iterable[ty.Union[str, Path]],
    colors: ty.Iterable[ty.Union[str, ty.Tuple[str, str]]],
    opacities: ty.Iterable[float] = (1.0,),
    theme_override: ty.Optional[ty.Dict[str, str]] = None,
):
    dest = Path(dest)
    dest.mkdir(parents=True, exist_ok=True)
    svgs = generate_colorized_svgs(
        svg_paths=svg_paths,
        colors=colors,
        opacities=opacities,
        theme_override=theme_override,
    )

    for alias, svg in svgs:
        (dest / Path(alias).name).write_text(svg)


def build_theme_svgs(theme_name: str) -> str:
    """Build theme SVGs."""
    from qtextra.assets import ICONS

    out = get_theme_cache_dir(theme_name)
    write_colorized_svgs(
        out,
        svg_paths=ICONS.values(),
        colors=[(theme_name, "icon")],
        opacities=(0.5, 1),
        theme_override={"warning": "warning", "logo_silhouette": "background"},
    )
    return str(out)


def is_dark() -> bool:
    """Check if theme is dark."""
    return THEMES.is_dark


THEMES: Themes = Themes()
THEMES.register_themes()
CANVAS: CanvasThemes = CanvasThemes()



"""Queue widget."""

from __future__ import annotations

import typing as ty
from contextlib import suppress

from loguru import logger
from qtpy.QtCore import Qt, Signal  # type: ignore[attr-defined]
from qtpy.QtWidgets import QScrollArea, QSizePolicy, QVBoxLayout, QWidget

import qtextra.helpers as hp
from qtextra.queue.cli_queue import CLIQueueHandler
from qtextra.queue.item import TaskWidget
from qtextra.queue.task import Task
from qtextra.typing import TaskState

logger = logger.bind(src="QueueWidget")


QUEUE = CLIQueueHandler()


class QueueList(QScrollArea):
    """Manager object."""

    MAX_TASKS: int = 500
    AUTO_EXPAND: bool = False

    evt_console = Signal(object)

    _hidden = False

    def __init__(self, parent: ty.Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.widgets: ty.Dict[str, TaskWidget] = {}
        self.finished_widgets: ty.List[TaskWidget] = []
        self.cancelled_widgets: ty.List[TaskWidget] = []
        self.failed_widgets: ty.List[TaskWidget] = []

        # states
        self.selected: ty.Sequence[str] = ["running", "queued", "paused", "finished", "failed", "cancelled"]

        # connect signals
        QUEUE.evt_queued.connect(self.on_task_queued)
        QUEUE.evt_started.connect(self.on_task_started)
        QUEUE.evt_paused.connect(self.on_task_paused)
        QUEUE.evt_next.connect(self.on_task_next)
        QUEUE.evt_finished.connect(self.on_task_finished)
        QUEUE.evt_errored.connect(self.on_task_failed)
        QUEUE.evt_part_errored.connect(self.on_task_part_failed)
        QUEUE.evt_cancelled.connect(self.on_task_cancelled)
        QUEUE.evt_progress.connect(self.on_task_progress)
        QUEUE.evt_remove_task.connect(self.on_remove_task)

        # setup UI
        scroll_widget = QWidget()  # type: ignore[unused-ignore]
        self.setWidget(scroll_widget)

        main_layout = QVBoxLayout(scroll_widget)
        main_layout.setSpacing(2)
        main_layout.setContentsMargins(1, 1, 1, 1)
        main_layout.addStretch(1)

        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)  # type: ignore[attr-defined]
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)  # type: ignore[attr-defined]
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)  # type: ignore[attr-defined]
        self._layout = main_layout

    def __repr__(self) -> str:
        """Repr."""
        return (
            f"{self.__class__.__name__}<{len(self.widgets)} widgets; {len(self.finished_widgets)} finished;"
            f" {len(self.cancelled_widgets)} cancelled; {len(self.failed_widgets)} failed>"
        )

    def widget_iter(self) -> ty.Iterable[TaskWidget]:
        """Iterate over widgets."""
        yield from self.widgets.values()
        yield from self.finished_widgets
        yield from self.cancelled_widgets
        yield from self.failed_widgets

    def task_iter(self) -> ty.Iterable[Task]:
        """Iterate over all tasks."""
        for widget in self.widget_iter():
            if widget.task:
                yield widget.task

    def validate(self, selected: ty.Optional[ty.Sequence[str]] = None) -> None:
        """Validate."""
        if selected is not None:
            self.selected = selected
        for widget in self.widgets.values():
            if widget.task:
                widget.setVisible(widget.task.state in self.selected)
        for widget in self.widget_iter():
            if widget.task:
                widget.setVisible(widget.task.state in self.selected)

    def purge_finished(self, max_tasks: ty.Optional[ty.Optional[int]] = None) -> None:
        """Purge tasks that are no longer necessary."""
        if max_tasks is None:
            max_tasks = self.MAX_TASKS
        if len(self.finished_widgets) > max_tasks:
            logger.trace(f"Purging finished tasks ({len(self.finished_widgets)}>{max_tasks}...")
            widgets = self.finished_widgets[max_tasks::]
            for widget in widgets:
                self._layout.removeWidget(widget)
                widget.deleteLater()
                del self.finished_widgets[self.finished_widgets.index(widget)]
                if widget.task:
                    logger.trace(f"Removed widget for '{widget.task.summary()}'")
        self.validate()

    def purge_all(self) -> None:
        """Remove all tasks."""
        logger.trace("Purging all tasks...")
        for widget in self.widget_iter():
            with suppress(RuntimeError):
                self._layout.removeWidget(widget)
            with suppress(RuntimeError):
                widget.deleteLater()
            with suppress(RuntimeError):
                if widget.task:
                    logger.trace(f"Removed widget for '{widget.task.summary()}'")
        self.widgets.clear()
        self.finished_widgets.clear()
        self.cancelled_widgets.clear()
        self.failed_widgets.clear()
        QUEUE.clear()
        self.validate()

    # # @Slot(Task)
    @staticmethod
    def on_start_task(task: Task) -> None:
        """Cancel task."""
        QUEUE.run_force(task)

    def on_requeue_task(self, task: Task) -> None:
        """Re-add task to the queue."""
        self._remove_widget(task)
        task.state = TaskState.QUEUED
        QUEUE.requeue(task, remove=True)

    def on_remove_task(self, task: str | Task) -> None:
        """Remove task from the queue."""
        self._remove_widget(task)
        QUEUE.remove(task)

    @staticmethod
    def on_check_task(task: Task) -> None:
        """Check whether task is in the queue."""
        QUEUE.check(task)

    @staticmethod
    def on_cancel_task(task: Task) -> None:
        """Cancel task."""
        QUEUE.cancel(task)

    @staticmethod
    def on_pause_task(task: Task, state: bool) -> None:
        """Cancel task."""
        QUEUE.pause(task, state)

    def _remove_widget(self, task: str | Task) -> None:
        # remove widget from the UI
        task_id = task.task_id if isinstance(task, Task) else task
        widget = self._find_widget(task)
        if widget:
            self._layout.removeWidget(widget)
            widget.deleteLater()
            widget.setParent(None)  # type: ignore[call-overload]
            if widget in self.finished_widgets:
                del self.finished_widgets[self.finished_widgets.index(widget)]
            if widget in self.failed_widgets:
                del self.failed_widgets[self.failed_widgets.index(widget)]
            if widget in self.cancelled_widgets:
                del self.cancelled_widgets[self.cancelled_widgets.index(widget)]
            self.widgets.pop(task_id, None)
            del widget
            logger.trace(f"Removed widget for '{task_id}'")

    def _find_widget(self, task: str | Task) -> ty.Optional[TaskWidget]:
        """Find widget by its task ID."""
        task_id = task.task_id if isinstance(task, Task) else task
        if task_id in self.widgets:
            return self.widgets[task_id]
        for widget in self.finished_widgets:
            if widget.task and widget.task.task_id == task_id:
                return widget
        for widget in self.cancelled_widgets:
            if widget.task and widget.task.task_id == task_id:
                return widget
        for widget in self.failed_widgets:
            if widget.task and widget.task.task_id == task_id:
                return widget
        return None

    def check_if_exists(self, task: Task) -> bool:
        """Check if task exists."""
        for task_ in self.task_iter():
            if task.task_id == task_.task_id:
                return True
        return False

    def on_task_queued(self, task: Task) -> None:
        """Task started."""
        if self.check_if_exists(task):
            logger.debug("Task already exists in the queue, skipping...")
            return

        task_id = task.task_id
        can_cancel, can_cancel_when_started = QUEUE.can_cancel(task)
        can_pause = QUEUE.can_pause(task)

        widget = TaskWidget()
        widget.set_task(
            task,
            can_cancel=can_cancel,
            can_pause=can_pause,
            can_cancel_when_started=can_cancel_when_started,
            can_force_start=QUEUE.CAN_FORCE_START,
            auto_expand=self.AUTO_EXPAND,
        )
        widget.evt_start_task.connect(self.on_start_task)
        widget.evt_cancel_task.connect(self.on_cancel_task)
        widget.evt_pause_task.connect(self.on_pause_task)
        widget.evt_requeue_task.connect(self.on_requeue_task)
        widget.evt_console.connect(self.evt_console.emit)  # type: ignore[unused-ignore]
        self._layout.insertWidget(0, widget)
        # if the task is already finished, add it to the finished list
        if task.state == TaskState.FINISHED:
            self.finished_widgets.append(widget)
        else:
            self.widgets[task_id] = widget
        self.purge_finished()

    def on_clear_queue(self, force: bool = False) -> None:
        """Clear table, but first ask for confirmation."""
        if force or hp.confirm(
            self, "Are you sure you wish to remove <b>all</b> tasks from the list?", "Clear queue..."
        ):
            self.purge_all()

    def on_requeue_failed(self) -> None:
        """Requeue all failed tasks."""
        tasks = [task.task for task in self.failed_widgets if task.task]
        for task in tasks:
            self.on_requeue_task(task)

    def on_requeue_cancelled(self) -> None:
        """Requeue all cancelled tasks."""
        tasks = [task.task for task in self.cancelled_widgets if task.task]
        for task in tasks:
            self.on_requeue_task(task)

    # @Slot(Task)
    def on_task_started(self, task: Task) -> None:
        """Task finished."""
        widget = self._find_widget(task)
        if widget:
            widget.started()
            self.validate()
        else:
            logger.warning(f"Could not find widget for task '{task.task_id}' (start)")

    def on_task_paused(self, task: Task, paused: bool) -> None:
        """Task finished."""
        widget = self._find_widget(task)
        if widget:
            widget.paused(paused)
            self.validate()
        else:
            logger.warning(f"Could not find widget for task '{task.task_id}' (paused)")

    # @Slot(Task)
    def on_task_next(self, task: Task) -> None:
        """Task finished."""
        widget = self._find_widget(task)
        if widget:
            widget.next()
            self.validate()
        else:
            logger.warning(f"Could not find widget for task '{task.task_id}' (next)")

    # @Slot(Task)
    def on_task_finished(self, task: Task) -> None:
        """Task finished."""
        widget = self.widgets.pop(task.task_id, None)
        if widget:
            widget.stop()
            self.finished_widgets.insert(1, widget)
        else:
            widget = self._find_widget(task)
            if widget:
                widget.stop()
            logger.warning(f"Could not find widget for task '{task.task_id}' (finished)")
        self.purge_finished()

    # @Slot(Task, tuple)
    def on_task_failed(self, task: Task, _exc_info: ty.Tuple) -> None:
        """Task failed."""
        widget = self.widgets.pop(task.task_id, None)
        if widget:
            widget.stop()
            self.failed_widgets.insert(1, widget)
        else:
            widget = self._find_widget(task)
            if widget:
                widget.stop()
            logger.warning(f"Could not find widget for task '{task.task_id}' (failed)")
        self.purge_finished()

    # @Slot(Task, tuple)
    def on_task_part_failed(self, task: Task, _exc_info: ty.Tuple) -> None:
        """Task failed."""
        widget = self._find_widget(task)
        if widget:
            widget.part_failed()

    # @Slot(Task)
    def on_task_cancelled(self, task: Task) -> None:
        """Task was cancelled."""
        widget = self.widgets.pop(task.task_id, None)
        if widget:
            widget.cancelled()
            self.cancelled_widgets.insert(1, widget)
        else:
            widget = self._find_widget(task)
            if widget:
                widget.cancelled()
            logger.warning(f"Could not find widget for task '{task.task_id}' (cancelled)")
        self.purge_finished()

    def on_task_progress(self, task: Task) -> None:
        """Task progressed."""
        widget = self._find_widget(task)
        if widget:
            widget.update_progress()



"""Popup for developer tools."""

from qtpy.QtWidgets import QLayout, QVBoxLayout, QWidget
from qtreload.qt_reload import QtReloadWidget

from qtextra.widgets.qt_dialog import QtFramelessTool


class QDevPopup(QtFramelessTool):
    HIDE_WHEN_CLOSE = True

    def __init__(self, parent: QWidget, modules: list[str]):
        self.modules = modules
        super().__init__(parent)
        self.setMinimumWidth(800)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QLayout:
        """Make panel."""
        self.qdev = QtReloadWidget(self.modules, self)

        _, hide_layout = self._make_hide_handle("Developer tools")
        layout = QVBoxLayout()
        layout.setSpacing(2)
        layout.setContentsMargins(2, 2, 2, 2)
        layout.addLayout(hide_layout)
        layout.addWidget(self.qdev, stretch=True)
        return layout


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import apply_style, qapplication

    app = qapplication(1)
    dlg = QDevPopup(None, ["qtextra"])
    apply_style(dlg)
    dlg.show()

    sys.exit(app.exec_())



"""System info dialog."""

from __future__ import annotations

from koyo.path import open_directory_alt
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QHBoxLayout, QLabel, QTextBrowser, QTextEdit, QVBoxLayout, QWidget

from qtextra.widgets.qt_button_clipboard import QtCopyToClipboardButton


class QtSystemInfo(QDialog):
    """Qt dialog window for displaying 'About {APP}}' information.

    Parameters
    ----------
    parent : QWidget, optional
        Parent of the dialog, to correctly inherit and apply theme.
        Default is None.

    Attributes
    ----------
    citationCopyButton : napari._qt.qt_about.QtCopyToClipboardButton
        Button to copy citation information to the clipboard.
    citationTextBox : qtpy.QtWidgets.QTextEdit
        Text box containing napari citation information.
    citation_layout : qtpy.QtWidgets.QHBoxLayout
        Layout widget for napari citation information.
    infoCopyButton : napari._qt.qt_about.QtCopyToClipboardButton
        Button to copy napari version information to the clipboard.
    info_layout : qtpy.QtWidgets.QHBoxLayout
        Layout widget for napari version information.
    infoTextBox : qtpy.QtWidgets.QTextEdit
        Text box containing napari version information.
    layout : qtpy.QtWidgets.QVBoxLayout
        Layout widget for the entire 'About napari' dialog.
    """

    def __init__(self, system_info: str, citation_info: str, title: str = "", parent: QWidget | None = None):
        super().__init__(parent)
        self.setWindowFlags(self.windowFlags() | Qt.WindowType.WindowStaysOnTopHint)

        self._layout = QVBoxLayout(self)
        self._layout.setSpacing(2)
        self._layout.setContentsMargins(5, 5, 5, 5)

        # Description
        title_label = QLabel(title)
        title_label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        self._layout.addWidget(title_label)

        # Add information
        self.infoTextBox = QTextBrowser(self)
        self.infoTextBox.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        self.infoTextBox.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
        self.infoTextBox.setOpenLinks(True)
        self.infoTextBox.setOpenExternalLinks(True)
        self.infoTextBox.anchorClicked.connect(open_directory_alt)

        # Add text copy button
        self.infoCopyButton = QtCopyToClipboardButton(self.infoTextBox)
        self.infoCopyButton.set_medium()
        self.info_layout = QHBoxLayout()
        self.info_layout.addWidget(self.infoTextBox, 1)
        self.info_layout.addWidget(self.infoCopyButton, 0, Qt.AlignmentFlag.AlignTop)
        self.info_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self._layout.addWidget(QLabel("<b>System information:</b>"))
        self._layout.addLayout(self.info_layout)

        self.infoTextBox.setText(system_info)
        self.infoTextBox.setMinimumSize(
            int(self.infoTextBox.document().size().width() + 19),
            int(min(self.infoTextBox.document().size().height() + 10, 500)),
        )

        self._layout.addWidget(QLabel("<b>Citation information:</b>"))
        self.citationTextBox = QTextEdit(citation_info)
        self.citationTextBox.setFixedHeight(64)
        self.citationCopyButton = QtCopyToClipboardButton(self.citationTextBox)
        self.citationCopyButton.set_medium()
        self.citation_layout = QHBoxLayout()
        self.citation_layout.addWidget(self.citationTextBox, 1)
        self.citation_layout.addWidget(self.citationCopyButton, 0, Qt.AlignmentFlag.AlignTop)
        self._layout.addLayout(self.citation_layout)

        self.resize(600, 400)

    @staticmethod
    def show_sys_info(
        system_info: str, citation_info: str, title: str = "System Information", parent: QWidget | None = None
    ) -> None:
        """Display the 'About napari' dialog box.

        Parameters
        ----------
        system_info : str
            Text containing system information.
        citation_info : str
            Text containing citation information.
        title : str
            Title of the dialog.
        parent : QWidget, optional
            Parent of the dialog, to correctly inherit and apply theme.
            Default is None.
        """
        d = QtSystemInfo(system_info, citation_info, title=title, parent=parent)
        d.setObjectName("DialogSystemInfo")
        d.setWindowTitle(title)
        d.setWindowModality(Qt.WindowModality.ApplicationModal)
        d.exec_()


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.config import THEMES
        from qtextra.utils.dev import qapplication

        _ = qapplication()  # analysis:ignore
        dlg = QtSystemInfo("TEST", "TEST", None)
        THEMES.set_theme_stylesheet(dlg)
        dlg.show()
        sys.exit(dlg.exec_())

    _main()  # type: ignore[no-untyped-call]



"""Tutorial widget."""

import typing as ty
from enum import Enum

from pydantic import BaseModel, ConfigDict, field_validator
from qtpy.QtCore import QEasingCurve, QPoint, Qt, QVariantAnimation
from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import QDialog, QGridLayout, QHBoxLayout, QProgressBar, QVBoxLayout, QWidget

import qtextra.helpers as hp


class Position(str, Enum):
    """Position."""

    CENTER = "center"

    LEFT_TOP = "left_top"
    LEFT = "left"
    LEFT_BOTTOM = "left_bottom"

    RIGHT_TOP = "right_top"
    RIGHT = "right"
    RIGHT_BOTTOM = "right_bottom"

    TOP_LEFT = "top_left"
    TOP = "top"
    TOP_RIGHT = "top_right"

    BOTTOM_LEFT = "bottom_left"
    BOTTOM = "bottom"
    BOTTOM_RIGHT = "bottom_right"


class TutorialStep(BaseModel):
    """Tutorial step.

    Attributes
    ----------
    title: str
        Title of the tutorial step.
    message: str
        Message to display. Can be HTML.
    widget: QWidget
        Widget to associate tutorial with. If widget is provided, chevron icon will be shown and it will point towards
        it.
    position: Position
        The position of the tutorial chevron.
    position_offset: tuple[int, int]
        Offset position of the tutorial chevron.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    title: str = ""
    message: str
    widget: QWidget
    position: Position = Position.RIGHT
    position_offset: tuple[int, int] = (0, 0)
    func: ty.Optional[tuple[ty.Callable, ...]] = None

    @field_validator("widget", mode="before")
    def validate_widget(widget: QWidget) -> QWidget:
        """Validate widget."""
        if not isinstance(widget, QWidget):
            raise ValueError(f"Invalid widget '{widget}'.")
        return widget


class QtTutorial(QDialog):
    """Tutorial step widget."""

    # Window attributes
    MIN_WIDTH = 350
    MAX_WIDTH = 450
    MIN_HEIGHT = 40
    ALLOW_CHEVRON = True

    _current = -1
    steps: ty.List[TutorialStep]
    chevrons: ty.Dict[str, ty.Optional[QWidget]]

    def __init__(self, parent: ty.Optional[QWidget] = None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.Tool)
        self.setSizeGripEnabled(False)
        self.setModal(False)

        self.setMinimumWidth(self.MIN_WIDTH)
        self.setMinimumHeight(self.MIN_HEIGHT)

        self._animation = QVariantAnimation()
        self._animation.setDuration(500)
        self._animation.valueChanged.connect(self._update_progress)
        self._animation.setEasingCurve(QEasingCurve.Type.InOutCubic)

        # self._move_animation = QVariantAnimation()
        # self._move_animation.setDuration(1000)
        # self._move_animation.valueChanged.connect(self._update_position)
        # self._move_animation.setEasingCurve(QEasingCurve.InOutCubic)

        self.steps = []
        self.make_ui()
        if not self.ALLOW_CHEVRON:
            for chevron in self.chevrons.values():
                chevron.hide()

    # noinspection PyAttributeOutsideInit
    def make_ui(self) -> None:
        """Setup UI."""
        self.chevron_up_left = hp.make_qta_label(self, "chevron_up_circle", small=True, retain_size=False)
        self.chevron_up_mid = hp.make_qta_label(self, "chevron_up_circle", small=True, retain_size=False)
        self.chevron_up_right = hp.make_qta_label(self, "chevron_up_circle", small=True, retain_size=False)

        self.chevron_down_left = hp.make_qta_label(self, "chevron_down_circle", small=True, retain_size=False)
        self.chevron_down_mid = hp.make_qta_label(self, "chevron_down_circle", small=True, retain_size=False)
        self.chevron_down_right = hp.make_qta_label(self, "chevron_down_circle", small=True, retain_size=False)

        self.chevron_left_top = hp.make_qta_label(self, "chevron_left_circle", small=True, retain_size=False)
        self.chevron_left_mid = hp.make_qta_label(self, "chevron_left_circle", small=True, retain_size=False)
        self.chevron_left_bottom = hp.make_qta_label(self, "chevron_left_circle", small=True, retain_size=False)

        self.chevron_right_top = hp.make_qta_label(self, "chevron_right_circle", small=True, retain_size=False)
        self.chevron_right_mid = hp.make_qta_label(self, "chevron_right_circle", small=True, retain_size=False)
        self.chevron_right_bottom = hp.make_qta_label(self, "chevron_right_circle", small=True, retain_size=False)

        self.chevrons = {
            Position.CENTER: None,
            Position.TOP_LEFT: self.chevron_up_left,
            Position.TOP: self.chevron_up_mid,
            Position.TOP_RIGHT: self.chevron_up_right,
            Position.BOTTOM_LEFT: self.chevron_down_left,
            Position.BOTTOM: self.chevron_down_mid,
            Position.BOTTOM_RIGHT: self.chevron_down_right,
            Position.LEFT_TOP: self.chevron_left_top,
            Position.LEFT: self.chevron_left_mid,
            Position.LEFT_BOTTOM: self.chevron_left_bottom,
            Position.RIGHT_TOP: self.chevron_right_top,
            Position.RIGHT: self.chevron_right_mid,
            Position.RIGHT_BOTTOM: self.chevron_right_bottom,
        }

        header_widget = QWidget(self)
        header_widget.setObjectName("tutorial_header")

        self._step_indicator = QProgressBar(header_widget)
        self._step_indicator.setObjectName("step_indicator")
        self._step_indicator.setTextVisible(False)
        self._close_btn = hp.make_qta_btn(
            header_widget, "cross", small=True, medium=False, func=self.close, tooltip="Close popup."
        )

        header_layout = QHBoxLayout(header_widget)
        header_layout.setContentsMargins(2, 2, 2, 2)
        header_layout.addWidget(self._step_indicator, stretch=True)
        header_layout.addWidget(self._close_btn)

        self._title_label = hp.make_label(self, "", bold=True)
        self._message_label = hp.make_label(self, "", wrap=True, selectable=True, enable_url=True)

        footer_widget = QWidget(self)
        footer_widget.setObjectName("tutorial_footer")
        self._step_label = hp.make_label(footer_widget, "", object_name="step_label")
        self._prev_btn = hp.make_btn(footer_widget, "Previous", func=self.on_prev, tooltip="Show previous step.")
        self._next_btn = hp.make_btn(footer_widget, "Next", func=self.on_next, tooltip="Show next step.")

        footer_layout = QHBoxLayout(footer_widget)
        footer_widget.setContentsMargins(2, 2, 2, 2)
        footer_layout.addWidget(self._step_label, stretch=True)
        footer_layout.addStretch(1)
        footer_layout.addWidget(self._prev_btn)
        footer_layout.addWidget(self._next_btn)

        # layout
        layout = QVBoxLayout()
        layout.setContentsMargins(4, 4, 4, 4)
        layout.setSpacing(0)
        layout.addWidget(header_widget)
        layout.addWidget(self._title_label)
        layout.addWidget(self._message_label, stretch=True)
        layout.addWidget(footer_widget)

        main_layout = QGridLayout(self)
        # widget, row, column, rowspan, colspan
        main_layout.setContentsMargins(2, 2, 2, 2)
        main_layout.setSpacing(0)
        main_layout.setColumnStretch(1, True)
        main_layout.setRowStretch(3, True)
        # top
        main_layout.addWidget(self.chevron_up_left, 0, 0, 1, 1, alignment=Qt.AlignmentFlag.AlignHCenter)
        main_layout.addWidget(self.chevron_up_mid, 0, 1, 1, -1, alignment=Qt.AlignmentFlag.AlignHCenter)
        main_layout.addWidget(self.chevron_up_right, 0, 2, 1, 1, alignment=Qt.AlignmentFlag.AlignHCenter)
        # content
        main_layout.addLayout(layout, 1, 1, 3, 1)
        # left
        main_layout.addWidget(self.chevron_left_top, 2, 0, 1, 1, alignment=Qt.AlignmentFlag.AlignVCenter)
        main_layout.addWidget(self.chevron_left_mid, 3, 0, 1, 1, alignment=Qt.AlignmentFlag.AlignVCenter)
        main_layout.addWidget(self.chevron_left_bottom, 4, 0, 1, 1, alignment=Qt.AlignmentFlag.AlignVCenter)
        # right
        main_layout.addWidget(self.chevron_right_top, 2, 2, 1, 1, alignment=Qt.AlignmentFlag.AlignVCenter)
        main_layout.addWidget(self.chevron_right_mid, 3, 2, 1, 1, alignment=Qt.AlignmentFlag.AlignVCenter)
        main_layout.addWidget(self.chevron_right_bottom, 4, 2, 1, 1, alignment=Qt.AlignmentFlag.AlignVCenter)
        # bottom
        main_layout.addWidget(self.chevron_down_left, 5, 0, 1, 1, alignment=Qt.AlignmentFlag.AlignHCenter)
        main_layout.addWidget(self.chevron_down_mid, 5, 1, 1, -1, alignment=Qt.AlignmentFlag.AlignHCenter)
        main_layout.addWidget(self.chevron_down_right, 5, 2, 1, 1, alignment=Qt.AlignmentFlag.AlignHCenter)

    def _update_progress(self, value: int) -> None:
        """Update progress bar."""
        self._step_indicator.setValue(value)

    def set_steps(self, steps: ty.List[TutorialStep]) -> None:
        """Set steps."""
        self.steps = steps
        self._step_indicator.setMinimum(0)
        self._step_indicator.setMaximum(len(steps) * 100)
        self.set_step(0)

    def set_step(self, index: int) -> None:
        """Show step."""
        self._current = index

        step = self.steps[index]
        if step.func:
            for func in step.func:
                func()
        self._title_label.setText(step.title)
        self._message_label.setText(step.message)
        self._step_label.setText(f"Step {index + 1}/{len(self.steps)}")

        # enable/disable buttons
        hp.disable_widgets(self._prev_btn, disabled=index == 0)
        self._next_btn.setText("Next" if index < len(self.steps) - 1 else "Done")
        # move tutorial to specified location
        self._message_label.adjustSize()
        self.adjustSize()
        self.set_chevron(step.position)
        self.move_to_widget(step.widget, step.position, step.position_offset)

        # update animation
        self._animation.setStartValue(self._step_indicator.value())
        self._animation.setEndValue((index + 1) * 100)
        self._animation.start()

    def move_to_widget(
        self, widget: QWidget, position: str = "right", position_offset: tuple[int, int] = (0, 0)
    ) -> None:
        """Move tutorial to specified widget."""
        position = Position(position)
        x_pad, y_pad = 5, 5
        popup_size = self.size()
        chevron = self.chevrons[position]
        if chevron:
            icon_pos = chevron.pos()
            x_offset = int(chevron.size().width() / 2)
            y_offset = int(chevron.size().height() / 2)
        else:
            icon_pos = QPoint(0, 0)
            x_offset = 0
            y_offset = 0

        x_pos_offset, y_pos_offset = position_offset
        rect_of_widget = widget.rect()
        if position in ["right", "right_top", "right_bottom"]:
            x = rect_of_widget.left() - popup_size.width() - x_pad - x_pos_offset
            y = rect_of_widget.center().y() - icon_pos.y() - y_offset - y_pos_offset
        elif position in ["left", "left_top", "left_bottom"]:
            x = rect_of_widget.right() + x_pad - x_pos_offset
            y = rect_of_widget.center().y() - icon_pos.y() - y_offset - y_pos_offset
        elif position in ["bottom", "bottom_left", "bottom_right"]:
            x = rect_of_widget.center().x() - icon_pos.x() - x_offset - x_pos_offset
            y = rect_of_widget.top() - popup_size.height() - y_pad - y_pos_offset
        elif position in ["top", "top_left", "top_right"]:
            x = rect_of_widget.center().x() - icon_pos.x() - x_offset - x_pos_offset
            y = rect_of_widget.bottom() + y_pad - y_pos_offset
        elif position in ["center"]:
            x = rect_of_widget.center().x() - popup_size.width() / 2
            y = rect_of_widget.center().y() - popup_size.height() / 2
        else:
            raise ValueError(f"Invalid position '{position}'.")
        pos = widget.mapToGlobal(QPoint(int(x), int(y)))
        pos = hp.check_if_outside_for_mouse(pos, popup_size)
        self.move(pos)

    def set_chevron(self, position: Position) -> None:
        """Show/hide chevron icons as required."""
        position = Position(position)
        if self.ALLOW_CHEVRON:
            for key, chevron in self.chevrons.items():
                if not chevron:
                    continue
                chevron.setVisible(key == position)

    def on_next(self) -> None:
        """Next step."""
        if self._current == len(self.steps) - 1:
            self.close()
        else:
            self.set_step(self._current + 1)

    def on_prev(self) -> None:
        """Previous step."""
        if self._current > 0:
            self.set_step(self._current - 1)

    def show(self) -> None:
        """Show widget."""
        if self._current == -1:
            self.on_next()
        super().show()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Key press event handler."""
        key = event.key()
        if key == Qt.Key.Key_Left:
            self.on_prev()
            event.accept()
        elif key == Qt.Key.Key_Right:
            self.on_next()
            event.accept()
        else:
            super().keyPressEvent(event)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    def _popover(frame, widget):
        text = """Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore
        et dolore magna aliqua. Vestibulum lorem sed risus ultricies tristique nulla aliquet. Malesuada nunc vel risus
         commodo viverra maecenas. Nascetur ridiculus mus mauris vitae ultricies leo. Tellus in hac habitasse platea
         dictumst vestibulum rhoncus. Egestas fringilla phasellus faucibus scelerisque eleifend donec pretium vulputate.
         Amet nulla facilisi morbi tempus iaculis urna id volutpat lacus. Aliquet nec ullamcorper sit amet risus nullam
         eget felis. Pharetra magna ac placerat vestibulum lectus. Dignissim convallis aenean et tortor at risus. Vitae
         tempus quam pellentesque nec nam aliquam sem et. Pulvinar proin gravida hendrerit lectus."""
        pop = QtTutorial(frame)
        pop.set_steps(
            [
                TutorialStep(
                    title=f"{position}",
                    message=text,
                    widget=widget,
                    position=position,
                )
                for position in Position
            ]
        )
        pop.show()

    app, frame, ha = qframe()

    _popover(frame, frame)
    frame.show()
    frame.setMaximumHeight(400)

    sys.exit(app.exec_())



from qtpy.QtCore import Qt
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QHBoxLayout, QLabel, QWidget

from qtextra.widgets.qt_separator import QtHorzLine


class QReadMoreLessLabel(QWidget):
    """Read more/less label."""

    def __init__(self, parent, text: str):
        QWidget.__init__(self, parent)

        self.text = text
        self.readmore = False

        # Explanation text
        self.explanation_layout = QHBoxLayout()
        self.explanation_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.mousePressEvent = self.state_toggle

        if "<moreless>" in text:
            self.readless_text, self.readmore_text = text.split("<moreless>")
            self.readmore_text = text
            self.readmore_left, self.readmore_right = self.readmore_text.split("<split>")
            text_left = self.readless_text + ("" if self.readmore_right.strip() == "" else "<b>Read more...</b>")
            text_right = ""
        else:
            text_left, text_right = text.split("<split>")

        self.explanation_text_left = QLabel(text_left, self)
        self.explanation_text_left.setWordWrap(True)
        self.explanation_text_left.setTextFormat(Qt.TextFormat.RichText)
        self.explanation_layout.addWidget(self.explanation_text_left, 50)

        # Vertical Line Break
        self.vertical_break = QtHorzLine(self)
        self.explanation_layout.addWidget(self.vertical_break)

        self.explanation_text_right = QLabel(text_right, self)
        self.explanation_text_right.setWordWrap(True)
        self.explanation_text_right.setTextFormat(Qt.TextFormat.RichText)
        self.explanation_layout.addWidget(self.explanation_text_right, 50)

        if self.readmore_text is not None and self.readmore_text.strip() == "":
            self.vertical_break.setHidden(True)
            self.explanation_text_right.setHidden(True)

        self.setLayout(self.explanation_layout)

    def state_toggle(self, a0: QMouseEvent) -> None:
        if self.readmore_text is not None and self.readmore_right.strip() != "":
            self.readmore = not self.readmore
            if self.readmore:
                self.explanation_text_left.setText(self.readmore_left)
                self.explanation_text_right.setText(self.readmore_right + "<b>Read less...</b>")
            else:
                self.explanation_text_left.setText(self.readless_text + "<b>Read more...</b>")
                self.explanation_text_right.setText("")


if __name__ == "__main__":  # pragma: no cover

    def _make_text():
        return """
    You can use this widget to create a label that can be expanded or collapsed.
    <br><br>
    Some info which can take a little bit of space
    <moreless>
    Some more text that will be hidden
    <split>
    Even some more text and even more text
    """

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)
        frame.setMinimumSize(600, 600)

        wdg = QReadMoreLessLabel(parent=frame, text=_make_text())
        ha.addWidget(wdg)

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Test qt_table_view"""

import pytest

from qtextra.widgets.qt_table_view_check import QtCheckableTableView, TableConfig


@pytest.fixture
def setup_table_widget(qtbot):
    """Setup panel"""

    def _widget() -> QtCheckableTableView:
        """Setup panel"""
        widget = QtCheckableTableView(None)
        qtbot.addWidget(widget)
        return widget

    return _widget


def test_widget_init(qtbot, setup_table_widget):
    widget = setup_table_widget()

    assert widget.n_rows == 0, "Widget should have 0 rows"
    assert widget.n_cols == 0, "Widget should have 0 columns"

    config = TableConfig().add("Test", "test").add("Test2", "test2")
    widget.setup_model_from_config(config)

    widget.add_row(["Test", "Test2"])
    assert widget.n_cols == 2, "Widget should have 2 columns"
    assert widget.n_rows == 1, "Widget should have 1 rows"
    widget.add_data([["Test", "Test2"]])
    assert widget.n_rows == 2, "Widget should have 2 rows"

    assert widget.get_value(0, 0) == "Test"
    assert widget.get_value(1, 0) == "Test2"
    assert widget.get_col_data(0) == ["Test", "Test"]
    assert widget.get_row_data(0) == ["Test", "Test2"]



"""Mini toolbar."""

from __future__ import annotations

import typing as ty

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QFrame, QHBoxLayout, QLayout, QVBoxLayout, QWidget

import qtextra.helpers as hp
from qtextra.typing import Orientation
from qtextra.widgets.qt_button_icon import QtImagePushButton


class QtMiniToolbar(QFrame):
    """Mini toolbar."""

    def __init__(
        self,
        parent: QWidget | None,
        orientation: Orientation | Qt.Orientation = Qt.Orientation.Horizontal,
        add_spacer: bool = True,
        icon_size: ty.Literal["small", "average", "medium", "normal"] | str | None = None,
        spacing: int = 0,
    ):
        super().__init__(parent)
        self._tools: dict[str, QtImagePushButton] = {}
        self.orientation = hp.get_orientation(orientation)

        self.layout_ = QHBoxLayout(self) if self.orientation == Qt.Orientation.Horizontal else QVBoxLayout(self)
        self.layout_.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.layout_.setSpacing(spacing)
        self.layout_.setContentsMargins(0, 0, 0, 0)
        if add_spacer:
            self.layout_.addSpacerItem(
                hp.make_h_spacer() if self.orientation == Qt.Orientation.Horizontal else hp.make_v_spacer()
            )

        self.max_size = 28
        self.icon_object_name, self.icon_size = (
            QtImagePushButton.get_icon_size_for_name(icon_size) if icon_size else None,
            None,
        )

    @property
    def max_size(self) -> int:
        """Return maximum size."""
        return self.maximumHeight() if self.orientation == Qt.Orientation.Horizontal else self.maximumWidth()

    @max_size.setter
    def max_size(self, value: int) -> None:
        current_max = self.max_size
        if current_max == value:
            return
        self.setMaximumHeight(value) if self.orientation == Qt.Orientation.Horizontal else self.setMaximumWidth(value)

    def _update_max_size(self, widget: QWidget, padding: int = 4):
        self.max_size = (
            widget.sizeHint().height() if self.orientation == Qt.Orientation.Horizontal else widget.sizeHint().width()
        ) + padding

    @property
    def n_items(self) -> int:
        """Return the number of items in the layout."""
        return self.layout_.count()

    def _make_qta_button(
        self,
        name: str,
        func: ty.Callable | None = None,
        func_menu: ty.Callable | None = None,
        tooltip: str | None = None,
        checkable: bool = False,
        check: bool = False,
        size: tuple[int, int] | None = None,
        flat: bool = False,
        small: bool = False,
        medium: bool = False,
        average: bool = False,
        normal: bool = False,
        checked_icon_name: ty.Optional[str] = None,
        object_name: str | None = None,
        hide: bool = False,
    ) -> QtImagePushButton:
        if self.icon_size:
            size = self.icon_size
            object_name = self.icon_object_name
        if not any((small, average, medium, normal)) and not size:
            size = (26, 26)
        if name in self._tools:
            raise ValueError(f"Tool '{name}' already exists.")
        btn = hp.make_qta_btn(
            self,
            name,
            tooltip=tooltip,
            flat=flat,
            medium=medium,
            size=size,
            checkable=checkable,
            checked=check,
            func=func,
            func_menu=func_menu,
            small=small,
            average=average,
            normal=normal,
            checked_icon_name=checked_icon_name,
            object_name=object_name,
            hide=hide,
        )
        self._tools[name] = btn
        return btn

    def add_qta_tool(
        self,
        name: str,
        func: ty.Callable | None = None,
        tooltip: str | None = None,
        checkable: bool = False,
        check: bool = False,
        size: tuple[int, int] | None = None,
        small: bool = False,
        average: bool = False,
        hide: bool = False,
    ) -> QtImagePushButton:
        """Insert tool."""
        btn = self._make_qta_button(
            name,
            func=func,
            tooltip=tooltip,
            checkable=checkable,
            check=check,
            size=size,
            small=small,
            average=average,
            hide=hide,
        )
        self.add_button(btn)
        return btn

    def add_layout(self, layout: QLayout) -> QLayout:
        """Insert any layout at specified position."""
        self.layout_.addLayout(layout)
        return layout

    def insert_layout(self, layout: QLayout, index: int = 0) -> QLayout:
        """Insert any layout at specified position."""
        self.layout_.insertLayout(index, layout)
        return layout

    def add_button(self, button: QtImagePushButton) -> QtImagePushButton:
        """Add any button the toolbar."""
        self.layout_.addWidget(button, alignment=Qt.AlignmentFlag.AlignCenter)
        self._update_max_size(button)
        return button

    def insert_button(self, button: QtImagePushButton, index: int = 0, set_size: bool = True) -> QtImagePushButton:
        """Insert any button at specified position."""
        if hasattr(button, "set_qta_size") and set_size:
            button.set_qta_size((26, 26))
        self.layout_.insertWidget(index, button, alignment=Qt.AlignmentFlag.AlignCenter)
        self._update_max_size(button)
        return button

    def add_widget(self, widget: QWidget, stretch: bool = False) -> QWidget:
        """Insert any widget at specified position."""
        kws = {}
        if not stretch:
            kws["alignment"] = Qt.AlignmentFlag.AlignCenter
        self.layout_.addWidget(widget, stretch=stretch, **kws)
        self._update_max_size(widget)
        return widget

    def insert_widget(self, widget: QWidget, index: int = 0) -> QWidget:
        """Insert any widget at specified position."""
        self.layout_.insertWidget(index, widget, alignment=Qt.AlignmentFlag.AlignCenter)
        self._update_max_size(widget)
        return widget

    def insert_qta_tool(
        self,
        name: str,
        index: int = 0,
        flat: bool = False,
        func: ty.Callable | None = None,
        func_menu: ty.Callable | None = None,
        tooltip: str | None = None,
        checkable: bool = False,
        check: bool = False,
        size: tuple[int, int] | None = None,
        small: bool = False,
        average: bool = False,
        normal: bool = False,
        hidden: bool = False,
        checked_icon_name: ty.Optional[str] = None,
    ) -> QtImagePushButton:
        """Insert tool."""
        btn = self._make_qta_button(
            name,
            flat=flat,
            func=func,
            func_menu=func_menu,
            tooltip=tooltip,
            checkable=checkable,
            check=check,
            size=size,
            small=small,
            average=average,
            normal=normal,
            checked_icon_name=checked_icon_name,
        )
        self.insert_button(btn, index)
        if hidden:
            btn.hide()
        return btn

    def add_separator(self) -> None:
        """Add separator."""
        sep = hp.make_v_line() if self.orientation == Qt.Orientation.Horizontal else hp.make_h_line(self)
        self.layout_.addWidget(sep)

    def insert_separator(self) -> None:
        """Insert horizontal or vertical separator."""
        sep = hp.make_v_line() if self.orientation == Qt.Orientation.Horizontal else hp.make_h_line(self)
        self.layout_.insertWidget(0, sep)

    def insert_spacer(self) -> None:
        """Insert spacer item."""
        spacer = (
            hp.make_spacer_widget()
        )  # make_v_spacer() if self.orientation == Qt.Orientation.Horizontal else make_h_spacer()
        self.layout_.insertWidget(0, spacer, stretch=True)

    def add_spacer(self) -> None:
        """Insert spacer item."""
        spacer = hp.make_spacer_widget()
        self.layout_.insertWidget(self.layout_.count(), spacer, stretch=True)

    def show_border(self) -> None:
        """Show border."""
        self.setFrameShape(QFrame.Shape.Box)

    def swap_orientation(self) -> None:
        """Swap orientation."""
        self.orientation = (
            QHBoxLayout.Direction.LeftToRight
            if self.orientation == Qt.Orientation.Vertical
            else QVBoxLayout.Direction.TopToBottom
        )
        self.layout_.setDirection(self.orientation)
        self.layout_.invalidate()
        self.layout_.update()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)

    wdg = QtMiniToolbar(None, orientation=Qt.Orientation.Horizontal)
    for icon in ["home", "settings", "help", "info", "warning", "error"]:
        wdg.add_qta_tool(icon, tooltip=icon, func=None)
    ha.addWidget(wdg)

    wdg = QtMiniToolbar(None, orientation=Qt.Orientation.Vertical)
    for icon in ["home", "settings", "help", "info", "warning", "error"]:
        wdg.add_qta_tool(icon, tooltip=icon, func=None)
    ha.addWidget(wdg)
    frame.show()
    sys.exit(app.exec_())



"""Path pytest-qt"""

import logging

import pytest
from koyo.system import IS_MAC
from loguru import logger

# os.environ["QTEXTRA_PYTEST"] = "1"
logger.enable("qtextra")
logger.remove()


class PropagateHandler(logging.Handler):
    def emit(self, record):
        logger = logging.getLogger(record.name)
        if logger.isEnabledFor(record.levelno):
            logger.handle(record)


@pytest.fixture(autouse=True, scope="session")
def cleanup_loguru():
    # Remove all handlers to make sure we don't interfere with tests.
    logger.remove()


@pytest.fixture(autouse=True, scope="session")
def propagate_loguru(cleanup_loguru):
    handler_id = logger.add(PropagateHandler(), format="{message}")
    yield
    logger.remove(handler_id)


if IS_MAC:
    # See: https://github.com/pytest-dev/pytest-qt/issues/521
    import pytestqt.plugin
    from pytestqt.plugin import qt_api

    def _process_events():
        """Calls app.processEvents() while taking care of capturing exceptions
        or not based on the given item's configuration.
        """
        app = qt_api.QtWidgets.QApplication.instance()
        if app is not None:
            # app.processEvents()
            pass

    pytestqt.plugin._process_events = _process_events



"""Table view of numpy or pandas array."""

from __future__ import annotations

import typing as ty

import numpy as np
import pandas as pd
from koyo.timer import MeasureTimer
from koyo.utilities import find_nearest_index_single
from loguru import logger
from qtpy.QtCore import QAbstractTableModel, QModelIndex, QRect, Qt, Signal  # type: ignore[attr-defined]
from qtpy.QtGui import QBrush, QColor, QKeyEvent
from qtpy.QtWidgets import QAbstractItemView, QHeaderView, QTableView, QWidget

from qtextra.utils.color import get_text_color

if ty.TYPE_CHECKING:
    from matplotlib.colors import Normalize

TEXT_COLOR: str = "#000000"
N_COLORS = 256
BATCH_SIZE = 50
INITIAL_SIZE = 100


class QtRotatedHeaderView(QHeaderView):
    """Horizontal header where the view is rotated by 90 degrees."""

    def __init__(self, parent=None):
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.setMinimumSectionSize(20)

    def paintSection(self, painter, rect, logicalIndex):
        """Paint section."""
        painter.save()
        # translate the painter such that rotate will rotate around the correct point
        painter.translate(rect.x() + rect.width(), rect.y())
        painter.rotate(90)
        # and have parent code paint at this location
        newrect = QRect(0, 0, rect.height(), rect.width())
        super().paintSection(painter, newrect, logicalIndex)
        painter.restore()

    def minimumSizeHint(self) -> Qt.SizeHint:
        """Minimum size hint."""
        size = super().minimumSizeHint()
        size.transpose()
        return size

    def sectionSizeFromContents(self, logicalIndex):
        """Section size from contents."""
        size = super().sectionSizeFromContents(logicalIndex)
        size.transpose()
        return size


class QtArrayTableModel(QAbstractTableModel):
    """Model for the table."""

    df: pd.DataFrame
    base_df: pd.DataFrame
    colors: dict[int, QColor] | None = None
    color_list: np.ndarray | None = None
    normalizer: Normalize | None = None
    max_color: QColor = None
    n_total: int = 0
    n_loaded: int = 0
    fmt: str = "{}"

    def __init__(self, parent: QWidget, data: ty.Union[np.ndarray, pd.DataFrame]):
        super().__init__(parent)
        self.set_data(data)

    def set_data(self, data: ty.Union[np.ndarray, pd.DataFrame]) -> None:
        """Set data in model."""
        if isinstance(data, np.ndarray):
            data = pd.DataFrame(data)
        assert data.ndim == 2, "The table can only display arrays with two-dimensions."
        self.df = data.iloc[:BATCH_SIZE, :]
        self.base_df = data
        self.colors = None
        self.color_list = None
        self.n_total = len(self.base_df)
        self.n_loaded = len(self.df)
        self.reset()

    def set_formatting(self, fmt: str) -> None:
        """Text formatter."""
        self.fmt = fmt

    def set_colormap(self, colormap: str, min_val: float | None = None, max_val: float | None = None) -> None:
        """Set colormap."""
        import matplotlib.cm
        import matplotlib.colors

        with MeasureTimer() as timer:
            if colormap:
                colormap = matplotlib.cm.get_cmap(colormap, lut=N_COLORS)
                if min_val is None:
                    min_val = self.base_df.min().min()
                if max_val is None:
                    max_val = self.base_df.max().max()
                normalizer = matplotlib.colors.Normalize(min_val, max_val, clip=True)

                colors = {}
                step_size = (abs(min_val) + abs(max_val)) / N_COLORS
                value_list = np.arange(min_val, max_val + step_size, step_size)
                # value_list = np.linspace(min_val, max_val, N_COLORS)
                for i, value in enumerate(value_list):
                    color = np.asarray(colormap(normalizer(value)))
                    colors[i] = QColor(*(255 * color).astype("int"))
                self.color_list = np.linspace(0, 1, N_COLORS)
                self.max_color = colors[i]
                self.colors = colors
                self.normalizer = normalizer
            else:
                self.colors, self.color_list = None, None
        logger.trace(f"Set colormap in {timer()}.")

    def reset(self) -> None:
        """Reset model."""
        self.beginResetModel()
        self.endResetModel()

    def reset_data(self) -> None:
        """Reset data."""
        with MeasureTimer() as timer:
            self.df = self.df.iloc[0:0]
            self.base_df = self.base_df.iloc[0:0]
            self.reset()
        logger.trace(f"Reset data in {timer()}.")

    def data(self, index: QModelIndex, role: Qt.ItemDataRole | None = None) -> ty.Any:
        """Parse data."""
        if not index.isValid():
            return None
        # background color
        if role == Qt.ItemDataRole.BackgroundRole:
            if self.colors and self.normalizer:
                value = self.normalizer(self.df.iloc[index.row(), index.column()])
                index = find_nearest_index_single(self.color_list, value)
                color = self.colors.get(index, self.max_color)
                return QBrush(color)
            return QBrush()
        # text color
        elif role == Qt.ItemDataRole.ForegroundRole:
            if self.colors and self.normalizer:
                value = self.normalizer(self.df.iloc[index.row(), index.column()])
                index = find_nearest_index_single(self.color_list, value)
                color = self.colors.get(index, self.max_color)
                return QBrush(get_text_color(color))
            return QBrush(QColor(TEXT_COLOR))
        # display value
        elif role == Qt.ItemDataRole.DisplayRole:
            value = self.df.iloc[index.row(), index.column()]
            return self.fmt.format(value)
        # check alignment role
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignCenter

    def headerData(
        self, index: QModelIndex, orientation: Qt.Orientation, role: Qt.ItemDataRole | None = None
    ) -> str | None:
        """Get header data."""
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return str(self.df.columns[index])
        if orientation == Qt.Orientation.Vertical and role == Qt.ItemDataRole.DisplayRole:
            return str(self.df.index[index])
        return None

    def rowCount(self, parent: QWidget | None = None, **kwargs: ty.Any) -> int:
        """Return number of rows."""
        return self.df.shape[0] if self.df is not None else 0

    def columnCount(self, parent: QWidget | None = None, **kwargs: ty.Any) -> int:
        """Return number of columns."""
        return self.df.shape[1] if self.df is not None else 0

    def canFetchMore(self, parent: QWidget | None = None) -> bool:
        """Check whether you can fetch more data."""
        return self.n_total >= self.n_loaded

    def fetchMore(self, index: QModelIndex) -> None:
        """Fetch more data."""
        with MeasureTimer() as timer:
            reminder = self.n_total - self.n_loaded
            items_to_fetch = min(reminder, BATCH_SIZE)
            self.beginInsertRows(QModelIndex(), self.n_loaded, self.n_loaded + items_to_fetch - 1)
            self.n_loaded += items_to_fetch
            self.df = self.base_df.iloc[: self.n_loaded, :]
            self.endInsertRows()
        logger.trace(f"Fetched {items_to_fetch} items in {timer()}.")

    def sort(self, column: int, order: Qt.SortOrder = ...) -> None:
        """Sort data."""
        self.beginResetModel()
        try:
            self.df = self.df.sort_values(self.df.columns[column], ascending=order == Qt.SortOrder.AscendingOrder)
        except TypeError:
            pass
        self.endResetModel()


class QtArrayTableView(QTableView):
    """Array table."""

    evt_key_release = Signal(QKeyEvent)

    model: ty.Callable[[], QtArrayTableModel]

    def __init__(self, *args: ty.Any, sortable: bool = False, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self.sortable = sortable
        if self.sortable:
            self.horizontalHeader().sectionClicked.connect(self.sortByColumn)
            self.horizontalHeader().setSortIndicatorShown(True)

    def sortByColumn(self, index: int) -> None:
        """Override method."""
        order = self.horizontalHeader().sortIndicatorOrder()
        return QTableView.sortByColumn(self, index, order)

    def keyReleaseEvent(self, event) -> None:
        """Process key event press."""
        super().keyReleaseEvent(event)
        row = self.currentIndex().row()
        col = self.currentIndex().column()
        event.row = lambda: row  # make row retrieval a function so its compatible with other methods
        event.column = lambda: col  # make row retrieval a function so its compatible with other methods
        self.evt_key_release.emit(event)

    def set_data(
        self,
        data: ty.Union[np.ndarray, pd.DataFrame],
        fmt: str = "{:d}",
        colormap: str | None = None,
        min_val: float | None = None,
        max_val: float | None = None,
    ) -> None:
        """Set data."""
        model = QtArrayTableModel(self, data)
        model.set_colormap(colormap, min_val, max_val)
        if fmt:
            model.set_formatting(fmt)
        self.setModel(model)
        self.init()

    def set_formatting(self, fmt: str) -> None:
        """Text formatter."""
        # let's perform simple test to make sure value can be rendered
        fmt.format(42.0)

        # set value on model
        model = self.model()
        if model:
            model.set_formatting(fmt)
            model.reset()

    def set_colormap(self, colormap: str, min_val: float | None = None, max_val: float | None = None) -> None:
        """Set colormap."""
        model: QtArrayTableModel = self.model()
        if model:
            model.set_colormap(colormap, min_val, max_val)
            model.reset()

    def init(self) -> None:
        """Initialize table to ensure correct visuals."""
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setDragEnabled(True)
        header = self.horizontalHeader()
        header.setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        # self.setHorizontalHeader(QtRotatedHeaderView(self))

    def reset_data(self) -> None:
        """Reset data."""
        model = self.model()
        if model:
            model.reset_data()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, va = qframe(False)
    frame.setMinimumSize(400, 400)

    table = QtArrayTableView()
    va.addWidget(table)
    table.set_data(
        np.asarray([[-1, 0, 1], [1, 0, -1]]),
        # np.random.randint(-255, 255, (5, 5)) / 255,
        fmt="{:.2f}",
        colormap="coolwarm",
        min_val=-1,
        max_val=1,
    )

    frame.show()
    sys.exit(app.exec_())



import pytest

from qtextra.widgets.qt_button import QtActivePushButton


@pytest.fixture
def setup_widget(qtbot):
    """Setup panel"""

    def _widget() -> QtActivePushButton:
        widget = QtActivePushButton("")
        qtbot.addWidget(widget)
        return widget

    return _widget


class TestQtActivePushButton:
    def test_init(self, setup_widget):
        widget = setup_widget()
        assert widget.active is False
        assert widget._pixmap is None
        widget.active = True
        assert widget.active is True
        assert widget._pixmap is not None



"""Task."""

from __future__ import annotations

from time import time as time_

from qtextra.typing import TaskState


class Task:
    """Task.

    Parameters
    ----------
    task_id : str
        Task ID.
    commands : list[list[str]]
        List of commands.
    task_name : str, optional
        Task name, by default "Task".
    state : TaskState, optional
        Task state, by default TaskState.QUEUED.
    task_name_repr: : str, optional
        Task title representation, by default None. If specified, it will be used in the UI as is instead of the 'hyper'
        representation.
    """

    task_id: str
    commands: list[list[str]]
    task_name: str
    task_name_repr: str | None = None
    state: TaskState
    stdout_data: list[str]
    stdout: str

    command_index: int = 0
    active: bool = False
    locked: bool = False
    start_time: float | None = None
    end_time: float | None = None

    def __init__(
        self,
        task_id: str,
        commands: list[list[str]],
        task_name: str = "Task",
        state: TaskState = TaskState.QUEUED,
        task_name_repr: str | None = None,
        task_name_tooltip: str | None = None,
    ):
        self.task_id = task_id
        self.task_name = task_name
        self.task_name_repr = task_name_repr
        self.task_name_tooltip = task_name_tooltip
        self.commands = commands
        self.state = state
        self.stdout = ""
        self.stdout_data = []

    @property
    def pretty_info(self) -> str:
        """Return pretty representation of the info."""
        return f"{self.command_index + 1}/{len(self.commands)} commands"

    @property
    def duration(self) -> float:
        """Return duration."""
        if self.start_time == 0 or self.start_time is None or self.end_time == 0 or self.end_time is None:
            return 0
        return self.end_time - self.start_time

    @property
    def current_duration(self) -> float:
        """Return duration at this moment in time."""
        if self.start_time == 0 or self.start_time is None:
            return 0
        return time_() - self.start_time

    def summary(self) -> str:
        """Summary."""
        return f"{self.task_name}: {self.state.name}"

    def command_args(self) -> list[list[str]]:
        """Command args."""
        return self.commands

    def command_iter(self, command_index: int = 0) -> list[str]:  # type: ignore[misc]
        """Command iterator."""
        if command_index:
            yield from self.commands[command_index:]
        else:
            yield from self.commands

    def is_active(self) -> bool:
        """Check if task is active."""
        return self.active

    def activate(self) -> None:
        """Activate task."""
        self.active = True

    def is_locked(self) -> bool:
        """Check if task is locked."""
        return self.locked

    def lock(self) -> None:
        """Lock task."""
        if self.end_time is None:
            self.end_time = time_()
        self.locked = False

    def set_command_index(self, index: int) -> None:
        """Set command index."""
        self.command_index = index

    @property
    def current_std_out(self) -> str:
        """Return current stdout."""
        if len(self.stdout_data) > 10:
            return "\n".join(self.stdout_data[-10:])
        return "\n".join(self.stdout_data)

    def append_output(self, stdout: str | None = None) -> None:
        """Append output."""
        if stdout:
            self.stdout_data.append(stdout)



"""Modified SystemTrayIcon."""

# Third-party imports
from qtpy.QtWidgets import QSystemTrayIcon


class QtSystemTrayIcon(QSystemTrayIcon):
    """Modified SystemTrayIcon."""

    is_destroyed = False

    def __init__(self, parent):
        super().__init__(parent)

    def deleteLater(self):
        """Override deleteLater method to set the state of the object."""
        self.is_destroyed = True
        super(QSystemTrayIcon, self).deleteLater()



"""Standard Qt button."""

from __future__ import annotations

import typing as ty

from qtpy.QtCore import QPointF, QSize, Qt, Signal
from qtpy.QtGui import QColor, QMovie, QPainter
from qtpy.QtWidgets import QHBoxLayout, QLabel, QPushButton, QSizePolicy, QWidget

from qtextra.config import THEMES


class QtPushButton(QPushButton):
    """Standard Qt button. Here to enable easier styling."""

    evt_right_click = Signal()
    has_right_click: bool = False

    def __init__(self, *args: ty.Any, text: str | None = None, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self._label = QLabel(self)
        self._layout = QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setLayout(self._layout)
        self._label.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self._label.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        self._label.setAlignment(Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter)
        self._label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self._layout.addWidget(self._label)

        if text is not None:
            self.setText(text)

    def setWordWrap(self, state: bool) -> None:
        """Set word wrap on the button."""
        label = self.findChild(QLabel)
        if label:
            label.setWordWrap(state)

    def setText(self, text: str) -> None:
        """Set text on the label."""
        self._label.setText(text)

    def setTextFormat(self, fm: Qt.TextFormat) -> None:
        """SEt text format."""
        self._label.setTextFormat(fm)

    def text(self) -> str:
        """Return text."""
        return self._label.text()

    def mousePressEvent(self, evt: QEvent) -> None:  # type: ignore[override]
        """Mouse press event."""
        if evt.button() == Qt.MouseButton.RightButton:  # type: ignore[attr-defined]
            self.evt_right_click.emit()
        else:
            super().mousePressEvent(evt)  # type: ignore[arg-type]

    def sizeHint(self) -> QSize:
        """Return size hints."""
        sh = super().sizeHint()
        lb_sh = self._label.sizeHint()
        width = max(sh.width(), lb_sh.width())
        height = max(sh.height(), lb_sh.height())
        sh.setWidth(width)
        sh.setHeight(height)
        return sh

    def paintEvent(self, event) -> None:
        """Paint event/."""
        super().paintEvent(event)
        painter = QPainter(self)
        if self.has_right_click:
            width = 4
            radius = 4
            x = self.rect().width() - (width * 2.0)
            y = self.rect().height() - (width * 2.0)
            color = THEMES.get_hex_color("success")
            painter.setPen(QColor(color))
            painter.setBrush(QColor(color))
            painter.drawEllipse(QPointF(x, y), radius, radius)

    def connect_to_right_click(self, func: ty.Callable) -> None:
        """Connect function right right-click.

        It is not possible to check whether a function is connected to a signal so its better to use this function to
        connect via this function which leaves behind a flag so the paint event will add rectangle to the edge so the
        user knows there is a right-click menu available.
        """
        from qtextra.helpers import set_properties

        self.evt_right_click.connect(func)
        self.has_right_click = True
        set_properties(self, {"right_click": True})


class QtActivePushButton(QtPushButton):
    """Qt button with activity indicator built-in."""

    _pixmap = None

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        from qtextra.helpers import make_gif

        self.loading_movie = make_gif("infinity")
        self.loading_movie.frameChanged.connect(self._update_icon)
        self.active = False

    def _update_icon(self, _frame: int) -> None:
        """Update frame."""
        self.setIcon(self.loading_movie.currentPixmap())

    @property
    def active(self) -> bool:
        """Update the state of the loading label."""
        return self.loading_movie.state() == QMovie.MovieState.Running  # type: ignore[no-any-return]

    @active.setter
    def active(self, value: bool) -> None:
        self.loading_movie.start() if value else self.loading_movie.stop()
        if not value:
            self.setIcon(None)

    def setIcon(self, icon):
        """Set icon."""
        self._pixmap = icon
        self.repaint()

    def paintEvent(self, event) -> None:
        """Paint event/."""
        super().paintEvent(event)
        painter = QPainter(self)
        if self._pixmap is not None:
            y = int((self.height() - self._pixmap.height()) / 2)
            painter.drawPixmap(5, y, self._pixmap)


class QtRichTextButton(QtPushButton):
    """Rich-text button."""

    def __init__(self, parent: QWidget | None = None, text: str | None = None):
        super().__init__(parent=parent, text=text)
        self._label.setTextFormat(Qt.TextFormat.RichText)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    def _test():
        btn2.active = not btn2.active

    def _test2():
        print("clicked")

    app, frame, ha = qframe(False)
    frame.setMinimumSize(600, 600)

    btn1 = QtPushButton(frame)
    btn1.clicked.connect(_test2)
    btn1.connect_to_right_click(_test2)
    btn1.setText("TEST STANDARD BUTTON")
    ha.addWidget(btn1)

    btn2 = QtActivePushButton(frame)
    btn2.clicked.connect(_test)
    btn2.setText("TEST ACTIVE BUTTON")
    ha.addWidget(btn2)

    btn3 = QtRichTextButton(frame)
    btn3.clicked.connect(_test2)
    btn3.setText("<span style='color:#ff00ff;'>TEST</span> <i>RICH</i> <b>button</b>")
    ha.addWidget(btn3)

    frame.show()
    sys.exit(app.exec_())



"""
Adapted from:
https://wiki.python.org/moin/PyQt/A%20full%20widget%20waiting%20indicator
Also from:
https://github.dev/royerlab/aydin/blob/c19595f37a163f6cd34243c5d5975cddb4a637c1/aydin/gui/_qt/custom_widgets/overlay.py.
"""

from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QBrush, QPainter, QPen
from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget

from qtextra.config import THEMES

# TODO: it's possible that QtActiveOverlay is not always centered


class QtActiveOverlay(QWidget):
    """Widget that displays that action is in progress."""

    timer = None
    counter: int = 0

    # Attributes
    REVERSE: bool = False
    N_DOTS: int = 5
    INTERVAL: int = 200
    SIZE = 20
    SPACING = 50

    def __init__(self, n_dots: int = 5, interval: int = 200, size: int = 20, spacing=50, parent=None):
        self.N_DOTS = n_dots
        self.INTERVAL = interval
        self.SIZE = size
        self.SPACING = spacing
        QWidget.__init__(self, parent)

    def paintEvent(self, event):
        """Paint event."""
        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Fill background
        painter.fillRect(event.rect(), QBrush(THEMES.get_qt_color("background")))

        # No border
        painter.setPen(QPen(Qt.PenStyle.NoPen))

        # Vertical position for the dots
        height = (self.height() - self.SIZE) // 2

        # Calculate total width of all dots with spacing
        total_width = (self.N_DOTS - 1) * self.SPACING + self.SIZE

        # Calculate left margin to center the dots horizontally
        horizontal_offset = (self.width() - total_width) // 2

        # Draw each dot
        for i in range(self.N_DOTS):
            if i <= self.counter:
                color = "success" if not self.REVERSE else "primary"
            else:
                color = "primary" if not self.REVERSE else "success"
            painter.setBrush(QBrush(THEMES.get_qt_color(color)))
            x = horizontal_offset + i * self.SPACING
            painter.drawEllipse(x, height, self.SIZE, self.SIZE)
        painter.end()

    def showEvent(self, event):
        """Show event."""
        self.timer = self.startTimer(self.INTERVAL)
        self.counter = 0

    def timerEvent(self, event):
        """Timer event."""
        self.counter += 1
        if self.counter >= self.N_DOTS:
            self.REVERSE = not self.REVERSE
            self.counter = 0
        self.update()

    def hideEvent(self, event):
        """Hide event."""
        self.killTimer(self.timer)
        self.hide()

    def sizeHint(self) -> QSize:
        """Return the size hint for the widget."""
        return QSize(self.SPACING * (self.N_DOTS + 1), 120)

    def minimumHeight(self) -> QSize:
        """Return the minimum height for the widget."""
        return QSize(self.SPACING * (self.N_DOTS + 1), 120)


class QtActiveWidget(QWidget):
    """Widget that displays activity."""

    def __init__(self, text: str = "", which: str = "infinity", size: tuple[int, int] = (64, 64), parent=None):
        super().__init__(parent)

        from qtextra.helpers import make_label, make_loading_gif

        label = make_label(self, text, bold=True)
        spinner, _ = make_loading_gif(self, size=size, which=which)

        layout = QVBoxLayout(self)
        layout.setSpacing(2)
        layout.addWidget(label, alignment=Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(spinner, alignment=Qt.AlignmentFlag.AlignCenter)


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, va = qframe(False)

        va.addWidget(QLabel("QtActiveWidget"), alignment=Qt.AlignmentFlag.AlignHCenter)
        va.addWidget(QtActiveWidget(parent=frame))
        va.addWidget(QLabel("QtActiveOverlay"), alignment=Qt.AlignmentFlag.AlignHCenter)
        va.addWidget(QtActiveOverlay(parent=frame))

        frame.show()
        sys.exit(app.exec_())

    _main()
